from autolens.data.fitting import fitting
from autolens.data.fitting.util import fitting_util
from autolens.model.galaxy import galaxy_data as gd

def fit_galaxy_data_with_galaxy(galaxy_datas, model_galaxy):
    """Given a *galaxy_data.GalaxyData* object, fit it using a *galaxy.Galaxy* instance.

    This factory automatically pairs the two depending on the quantity being fitted.

    Parameters
    ----------
    galaxy_datas : [GalaxyData]
        The galaxy-data objects being fitted.
    model_galaxy : galaxy.Galaxy
        The model galaxy used to fit the galaxy-data.
    """
    if isinstance(galaxy_datas[0], gd.GalaxyDataIntensities) or isinstance(galaxy_datas[0], gd.GalaxyDataSurfaceDensity) or \
        isinstance(galaxy_datas[0], gd.GalaxyDataPotential):
        return GalaxyFit(galaxy_datas=galaxy_datas, model_galaxy=model_galaxy)

    elif isinstance(galaxy_datas[0], gd.GalaxyDataDeflectionsY) or isinstance(galaxy_datas[0], gd.GalaxyDataDeflectionsX):
        return GalaxyFitDeflections(galaxy_datas=galaxy_datas, model_galaxy=model_galaxy)

def fast_likelihood_from_galaxy_data_and_galaxy(galaxy_datas, model_galaxy):
    """Given a *galaxy_data.GalaxyData* object, fit it using a *model_galaxy.Galaxy* instance.

    The likelihood is computed in the fastest, least memory-intensive, way possible, for efficient non-linear sampling.

    This factory automatically pairs the two depending on the quantity being fitted.

    Parameters
    ----------
    galaxy_datas : [GalaxyData]
        The model_galaxy-data objects being fitted.
    model_galaxy : model_galaxy.Galaxy
        The model model_galaxy used to fit the model_galaxy-data.
    """
    if isinstance(galaxy_datas[0], gd.GalaxyDataIntensities) or isinstance(galaxy_datas[0], gd.GalaxyDataSurfaceDensity) or \
        isinstance(galaxy_datas[0], gd.GalaxyDataPotential):
        return GalaxyFit.fast_likelihood(galaxy_datas=galaxy_datas, galaxy=model_galaxy)

    elif isinstance(galaxy_datas[0], gd.GalaxyDataDeflectionsY) or isinstance(galaxy_datas[0], gd.GalaxyDataDeflectionsX):
        return GalaxyFitDeflections.fast_likelihood(galaxy_datas=galaxy_datas, model_galaxy=model_galaxy)


class GalaxyFit(fitting.AbstractDataFit):

    def __init__(self, galaxy_datas, model_galaxy):
        """Class which fits a set of galaxy-datas to a model galaxy, using either the galaxy's intensities, \
        surface-density or potential.

        Parameters
        ----------
        galaxy_datas : [GalaxyData]
            The galaxy-data object being fitted.
        model_galaxy : galaxy.Galaxy
            The model galaxy used to fit the galaxy-data.
        """

        self.galaxy_datas = galaxy_datas
        self.model_galaxy = model_galaxy

        model_data_ = galaxy_datas[0].profile_quantity_from_galaxy_and_sub_grid(galaxy=model_galaxy,
                                                                            sub_grid=galaxy_datas[0].grids.sub)

        super(GalaxyFit, self).__init__(fitting_datas=galaxy_datas, model_datas_=[model_data_])

    @classmethod
    def fast_likelihood(cls, galaxy_datas, galaxy):
        """Perform the fit of this class as described above, but storing no results as class instances, thereby \
        minimizing memory use and maximizing run-speed."""
        model_datas_ = galaxy_datas[0].profile_quantity_from_galaxy_and_sub_grid(galaxy=galaxy,
                                                                             sub_grid=galaxy_datas[0].grids.sub)
        residuals_ = fitting_util.residuals_from_datas_and_model_datas([galaxy_datas[:]], [model_datas_])
        chi_squareds_ = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals_, [galaxy_datas[0].noise_map_])
        chi_squared_terms = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds_)
        noise_terms = fitting_util.noise_terms_from_noise_maps([galaxy_datas[0].noise_map_])
        return sum(fitting_util.likelihoods_from_chi_squareds_and_noise_terms(chi_squared_terms, noise_terms))


class GalaxyFitDeflections(fitting.AbstractDataFit):

    def __init__(self, galaxy_datas, model_galaxy):
        """Class which fits a set of galaxy-datas to a model galaxy, using the galaxy's deflection-angle maps.

        Parameters
        ----------
        galaxy_datas : [GalaxyData]
            The galaxy-data object being fitted.
        model_galaxy : galaxy.Galaxy
            The model galaxy used to fit the galaxy-data.
        """
        self.galaxy_datas = galaxy_datas
        self.model_galaxy = model_galaxy

        model_data_ = galaxy_datas[0].profile_quantity_from_galaxy_and_sub_grid(galaxy=model_galaxy,
                                                                            sub_grid=galaxy_datas[0].grids.sub)

        super(GalaxyFitDeflections, self).__init__(fitting_datas=galaxy_datas, model_datas_=[model_data_[:, 0],
                                                                                              model_data_[:,1]])

    @classmethod
    def fast_likelihood(cls, galaxy_datas, model_galaxy):
        """Perform the fit of this class as described above, but storing no results as class instances, thereby \
        minimizing memory use and maximizing run-speed."""
        model_datas_ = galaxy_datas[0].profile_quantity_from_galaxy_and_sub_grid(galaxy=model_galaxy,
                                                                              sub_grid=galaxy_datas[0].grids.sub)

        residuals_ = fitting_util.residuals_from_datas_and_model_datas(galaxy_datas, [model_datas_[:, 0],
                                                                                 model_datas_[:,1]])
        chi_squareds_ = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals_, [galaxy_datas[0].noise_map_,
                                                                                        galaxy_datas[0].noise_map_])
        chi_squared_terms = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds_)
        noise_terms = fitting_util.noise_terms_from_noise_maps([galaxy_datas[0].noise_map_, galaxy_datas[0].noise_map_])
        return sum(fitting_util.likelihoods_from_chi_squareds_and_noise_terms(chi_squared_terms, noise_terms))


import numpy as np
import pytest
import scipy.spatial

from autolens.data.array import grids, mask
from autolens.model.inversion.util import pixelization_util
from autolens.model.inversion import regularization
from autolens.model.galaxy import galaxy as g


class TestRectangular:

    class TestNeighbors:

        def test__3x3_grid(self):

            # |0|1|2|
            # |3|4|5|
            # |6|7|8|

            pixel_neighbors, pixel_neighbors_size = pixelization_util.rectangular_neighbors_from_shape(shape=(3, 3))

            assert (pixel_neighbors[0] == [1, 3, -1, -1]).all()
            assert (pixel_neighbors[1] == [0, 2, 4, -1]).all()
            assert (pixel_neighbors[2] == [1, 5, -1, -1]).all()
            assert (pixel_neighbors[3] == [0, 4, 6, -1]).all()
            assert (pixel_neighbors[4] == [1, 3, 5, 7]).all()
            assert (pixel_neighbors[5] == [2, 4, 8, -1]).all()
            assert (pixel_neighbors[6] == [3, 7, -1, -1]).all()
            assert (pixel_neighbors[7] == [4, 6, 8, -1]).all()
            assert (pixel_neighbors[8] == [5, 7, -1, -1]).all()

            assert (pixel_neighbors_size == np.array([2, 3, 2, 3, 4, 3, 2, 3, 2])).all()

        def test__3x4_grid(self):

            # |0|1| 2| 3|
            # |4|5| 6| 7|
            # |8|9|10|11|

            pixel_neighbors, pixel_neighbors_size = pixelization_util.rectangular_neighbors_from_shape(shape=(3, 4))

            assert (pixel_neighbors[0] == [1, 4, -1, -1]).all()
            assert (pixel_neighbors[1] == [0, 2, 5, -1]).all()
            assert (pixel_neighbors[2] == [1, 3, 6, -1]).all()
            assert (pixel_neighbors[3] == [2, 7, -1, -1]).all()
            assert (pixel_neighbors[4] == [0, 5, 8, -1]).all()
            assert (pixel_neighbors[5] == [1, 4, 6, 9]).all()
            assert (pixel_neighbors[6] == [2, 5, 7, 10]).all()
            assert (pixel_neighbors[7] == [3, 6, 11, -1]).all()
            assert (pixel_neighbors[8] == [4, 9, -1, -1]).all()
            assert (pixel_neighbors[9] == [5, 8, 10, -1]).all()
            assert (pixel_neighbors[10] == [6, 9, 11, -1]).all()
            assert (pixel_neighbors[11] == [7, 10, -1, -1]).all()

            assert (pixel_neighbors_size == np.array([2, 3, 3, 2, 3, 4, 4, 3, 2, 3, 3, 2])).all()

        def test__4x3_grid(self):

            # |0| 1| 2|
            # |3| 4| 5|
            # |6| 7| 8|
            # |9|10|11|

            pixel_neighbors, pixel_neighbors_size = pixelization_util.rectangular_neighbors_from_shape(shape=(4, 3))

            assert (pixel_neighbors[0] == [1, 3, -1, -1]).all()
            assert (pixel_neighbors[1] == [0, 2, 4, -1]).all()
            assert (pixel_neighbors[2] == [1, 5, -1, -1]).all()
            assert (pixel_neighbors[3] == [0, 4, 6, -1]).all()
            assert (pixel_neighbors[4] == [1, 3, 5, 7]).all()
            assert (pixel_neighbors[5] == [2, 4, 8, -1]).all()
            assert (pixel_neighbors[6] == [3, 7, 9, -1]).all()
            assert (pixel_neighbors[7] == [4, 6, 8, 10]).all()
            assert (pixel_neighbors[8] == [5, 7, 11, -1]).all()
            assert (pixel_neighbors[9] == [6, 10, -1, -1]).all()
            assert (pixel_neighbors[10] == [7, 9, 11, -1]).all()
            assert (pixel_neighbors[11] == [8, 10, -1, -1]).all()

            assert (pixel_neighbors_size == np.array([2, 3, 2, 3, 4, 3, 3, 4, 3, 2, 3, 2])).all()

        def test__4x4_grid(self):

            # |0 | 1| 2| 3|
            # |4 | 5| 6| 7|
            # |8 | 9|10|11|
            # |12|13|14|15|

            pixel_neighbors, pixel_neighbors_size = pixelization_util.rectangular_neighbors_from_shape(shape=(4, 4))

            assert (pixel_neighbors[0] == [1, 4, -1, -1]).all()
            assert (pixel_neighbors[1] == [0, 2, 5, -1]).all()
            assert (pixel_neighbors[2] == [1, 3, 6, -1]).all()
            assert (pixel_neighbors[3] == [2, 7, -1, -1]).all()
            assert (pixel_neighbors[4] == [0, 5, 8, -1]).all()
            assert (pixel_neighbors[5] == [1, 4, 6, 9]).all()
            assert (pixel_neighbors[6] == [2, 5, 7, 10]).all()
            assert (pixel_neighbors[7] == [3, 6, 11, -1]).all()
            assert (pixel_neighbors[8] == [4, 9, 12, -1]).all()
            assert (pixel_neighbors[9] == [5, 8, 10, 13]).all()
            assert (pixel_neighbors[10] == [6, 9, 11, 14]).all()
            assert (pixel_neighbors[11] == [7, 10, 15, -1]).all()
            assert (pixel_neighbors[12] == [8, 13, -1, -1]).all()
            assert (pixel_neighbors[13] == [9, 12, 14, -1]).all()
            assert (pixel_neighbors[14] == [10, 13, 15, -1]).all()
            assert (pixel_neighbors[15] == [11, 14, -1, -1]).all()

            assert (pixel_neighbors_size == np.array([2, 3, 3, 2, 3, 4, 4, 3, 3, 4, 4, 3, 2, 3, 3, 2])).all()


class TestVoronoi:

    class TestNeighbors:

        def test__points_in_x_cross_shape__neighbors_of_each_pixel_correct(self):

            points = np.array([[1.0, -1.0], [1.0, 1.0],
                               [0.0, 0.0],
                               [-1.0, -1.0], [-1.0, 1.0]])

            voronoi = scipy.spatial.Voronoi(points, qhull_options='Qbb Qc Qx Qm')
            pixel_neighbors, pixel_neighbors_size = pixelization_util.voronoi_neighbors_from_pixels_and_ridge_points(
                pixels=5, ridge_points=np.array(voronoi.ridge_points))

            assert set(pixel_neighbors[0]) == {1, 2, 3, -1}
            assert set(pixel_neighbors[1]) == {0, 2, 4, -1}
            assert set(pixel_neighbors[2]) == {0, 1, 3, 4}
            assert set(pixel_neighbors[3]) == {0, 2, 4, -1}
            assert set(pixel_neighbors[4]) == {1, 2, 3, -1}

            assert (pixel_neighbors_size == np.array([3, 3, 4, 3, 3])).all()

        def test__9_points_in_square___neighbors_of_each_pixel_correct(self):

            # 9 points in a square - makes a square (this is the example int he scipy documentaiton page)

            points = np.array([[2.0, 0.0], [2.0, 1.0], [2.0, 2.0],
                               [1.0, 0.0], [1.0, 1.0], [1.0, 2.0],
                               [0.0, 0.0], [0.0, 1.0], [0.0, 2.0]])

            voronoi = scipy.spatial.Voronoi(points, qhull_options='Qbb Qc Qx Qm')
            pixel_neighbors, pixel_neighbors_size = pixelization_util.voronoi_neighbors_from_pixels_and_ridge_points(
                pixels=9, ridge_points=np.array(voronoi.ridge_points))

            assert set(pixel_neighbors[0]) == {1, 3, -1, -1}
            assert set(pixel_neighbors[1]) == {0, 2, 4, -1}
            assert set(pixel_neighbors[2]) == {1, 5, -1, -1}
            assert set(pixel_neighbors[3]) == {0, 4, 6, -1}
            assert set(pixel_neighbors[4]) == {1, 3, 5, 7}
            assert set(pixel_neighbors[5]) == {2, 4, 8, -1}
            assert set(pixel_neighbors[6]) == {3, 7, -1, -1}
            assert set(pixel_neighbors[7]) == {4, 6, 8, -1}
            assert set(pixel_neighbors[8]) == {5, 7, -1, -1}

            assert (pixel_neighbors_size == np.array([2, 3, 2, 3, 4, 3, 2, 3, 2])).all()


import os
import shutil

from autofit import conf
from autofit.core import non_linear as nl
from autolens.model.galaxy import galaxy, galaxy_model as gm
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp
from test.integration import tools

dirpath = os.path.dirname(os.path.realpath(__file__))
dirpath = os.path.dirname(dirpath)
output_path = '{}/../output/model_mapper'.format(dirpath)

path = '{}/../../'.format(os.path.dirname(os.path.realpath(__file__)))
conf.instance = conf.Config(config_path=path+'config', output_path=path+'output')

def pipeline():

    pipeline_name = "set_constant_float"
    data_name = '/set_constant_float'

    tools.reset_paths(data_name, pipeline_name, output_path)

    sersic = lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, intensity=1.0, effective_radius=1.3,
                                 sersic_index=3.0)

    lens_galaxy = galaxy.Galaxy(light_profile=sersic)

    tools.simulate_integration_image(data_name=data_name, pixel_scale=0.5, lens_galaxies=[lens_galaxy],
                                     source_galaxies=[], target_signal_to_noise=10.0)
    image = tools.load_image(data_name=data_name, pixel_scale=0.5)

    cf_pipeline = make_pipeline(pipeline_name=pipeline_name)

    results = cf_pipeline.run(image=image)
    for result in results:
        print(result)


def make_pipeline(pipeline_name):
    class MMPhase(ph.LensPlanePhase):

        def pass_priors(self, previous_results):
            self.lens_galaxies[0].sersic.axis_ratio = 0.2
            self.lens_galaxies[0].sersic.phi = 90.0
            self.lens_galaxies[0].sersic.intensity = 1.0
            self.lens_galaxies[0].sersic.effective_radius = 1.3
            self.lens_galaxies[0].sersic.sersic_index = 3.0

    phase1 = MMPhase(lens_galaxies=[gm.GalaxyModel(sersic=lp.EllipticalSersic)],
                     optimizer_class=nl.MultiNest, phase_name="{}/phase1".format(pipeline_name))

    phase1.optimizer.n_live_points = 20
    phase1.optimizer.sampling_efficiency = 0.8

    return pl.PipelineImaging(pipeline_name, phase1)


if __name__ == "__main__":
    pipeline()



import logging

logger = logging.getLogger(__name__)


class Pipeline(object):

    def __init__(self, pipeline_name, *phases):
        """

        Parameters
        ----------
        pipeline_name: str
            The phase_name of this pipeline
        """
        self.pipeline_name = pipeline_name
        self.phases = phases

    def __add__(self, other):
        """
        Compose two runners

        Parameters
        ----------
        other: Pipeline
            Another pipeline

        Returns
        -------
        composed_pipeline: Pipeline
            A pipeline that runs all the  phases from this pipeline and then all the phases from the other pipeline
        """
        return self.__class__("{} + {}".format(self.pipeline_name, other.pipeline_name), *(self.phases + other.phases))


class PipelineImaging(Pipeline):

    def __init__(self, pipeline_name, *phases):
        super(PipelineImaging, self).__init__(pipeline_name, *phases)

    def run(self, image, mask=None):

        from autolens.pipeline import phase as ph
        results = []
        for i, phase in enumerate(self.phases):
            logger.info("Running Phase {} (Number {})".format(phase.phase_name, i))
            if isinstance(phase, ph.HyperOnly):
                results[-1].hyper = phase.hyper_run(image, ph.ResultsCollection(results), mask)
            else:
                results.append(phase.run(image, ph.ResultsCollection(results), mask))
        return results


class PipelinePositions(Pipeline):

    def __init__(self, pipeline_name, *phases):
        super(PipelinePositions, self).__init__(pipeline_name, *phases)

    def run(self, positions, pixel_scale):
        from autolens.pipeline import phase as ph

        results = []
        for i, phase in enumerate(self.phases):
            logger.info("Running Phase {} (Number {})".format(phase.phase_name, i))
            results.append(phase.run(positions, pixel_scale, ph.ResultsCollection(results)))
        return results



import numpy as np
import pytest

from autolens import exc
from autolens.data.array import grids, mask
from autolens.model.inversion.util import inversion_util

class TestDataVectorFromData(object):

    def test__simple_blurred_mapping_matrix__correct_data_vector(self):

        blurred_mapping_matrix = np.array([[1.0, 1.0, 0.0],
                                           [1.0, 0.0, 0.0],
                                           [0.0, 1.0, 0.0],
                                           [0.0, 1.0, 1.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0]])

        image = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
        noise_map = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        data_vector = inversion_util.data_vector_from_blurred_mapping_matrix_and_data(
            blurred_mapping_matrix=blurred_mapping_matrix, image=image, noise_map=noise_map)

        assert (data_vector == np.array([2.0, 3.0, 1.0])).all()

    def test__simple_blurred_mapping_matrix__change_image_values__correct_data_vector(self):

        blurred_mapping_matrix = np.array([[1.0, 1.0, 0.0],
                                           [1.0, 0.0, 0.0],
                                           [0.0, 1.0, 0.0],
                                           [0.0, 1.0, 1.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0]])

        image = np.array([3.0, 1.0, 1.0, 10.0, 1.0, 1.0])
        noise_map = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        data_vector = inversion_util.data_vector_from_blurred_mapping_matrix_and_data(
            blurred_mapping_matrix=blurred_mapping_matrix, image=image, noise_map=noise_map)

        assert (data_vector == np.array([4.0, 14.0, 10.0])).all()

    def test__simple_blurred_mapping_matrix__change_noise_values__correct_data_vector(self):

        blurred_mapping_matrix = np.array([[1.0, 1.0, 0.0],
                                           [1.0, 0.0, 0.0],
                                           [0.0, 1.0, 0.0],
                                           [0.0, 1.0, 1.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0]])

        image = np.array([4.0, 1.0, 1.0, 16.0, 1.0, 1.0])
        noise_map = np.array([2.0, 1.0, 1.0, 4.0, 1.0, 1.0])

        data_vector = inversion_util.data_vector_from_blurred_mapping_matrix_and_data(
            blurred_mapping_matrix=blurred_mapping_matrix, image=image, noise_map=noise_map)

        assert (data_vector == np.array([2.0, 3.0, 1.0])).all()
        

class TestCurvatureMatrixFromBlurred(object):

    def test__simple_blurred_mapping_matrix(self):

        blurred_mapping_matrix = np.array([[1.0, 1.0, 0.0],
                                           [1.0, 0.0, 0.0],
                                           [0.0, 1.0, 0.0],
                                           [0.0, 1.0, 1.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0]])

        noise_map = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        curvature_matrix = inversion_util.curvature_matrix_from_blurred_mapping_matrix(
            blurred_mapping_matrix=blurred_mapping_matrix, noise_map=noise_map)

        assert (curvature_matrix == np.array([[2.0, 1.0, 0.0],
                                              [1.0, 3.0, 1.0],
                                              [0.0, 1.0, 1.0]])).all()

    def test__simple_blurred_mapping_matrix__change_noise_values(self):

        blurred_mapping_matrix = np.array([[1.0, 1.0, 0.0],
                                           [1.0, 0.0, 0.0],
                                           [0.0, 1.0, 0.0],
                                           [0.0, 1.0, 1.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0]])

        noise_map = np.array([2.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        curvature_matrix = inversion_util.curvature_matrix_from_blurred_mapping_matrix(
            blurred_mapping_matrix=blurred_mapping_matrix, noise_map=noise_map)

        assert (curvature_matrix == np.array([[1.25, 0.25, 0.0],
                                              [0.25, 2.25, 1.0],
                                              [0.0, 1.0, 1.0]])).all()


import os
import shutil

import numpy as np
import pytest

from autofit import conf
from autolens.data.fitting import fitting_data, fitting
from autolens.model.galaxy import galaxy as g
from autolens.data.imaging import image as im
from autolens.data.array import mask as msk, scaled_array
from autolens.data.fitting.plotters import fitting_plotters
from autolens.model.profiles import light_profiles as lp


@pytest.fixture(name='general_config')
def test_general_config():
    general_config_path = "{}/../../../test_files/configs/plotting/".format(os.path.dirname(os.path.realpath(__file__)))
    conf.instance.general = conf.NamedConfig(general_config_path + "general.ini")


@pytest.fixture(name='fitting_plotter_path')
def test_fitting_plotter_setup():
    galaxy_plotter_path = "{}/../../../test_files/plotting/fitting/".format(os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(galaxy_plotter_path):
        shutil.rmtree(galaxy_plotter_path)

    os.mkdir(galaxy_plotter_path)

    return galaxy_plotter_path


@pytest.fixture(name='image')
def test_image():
    image = scaled_array.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=1.0)
    noise_map = im.NoiseMap(array=2.0 * np.ones((3, 3)), pixel_scale=1.0)
    psf = im.PSF(array=3.0 * np.ones((1, 1)), pixel_scale=1.0)

    return im.Image(array=image, pixel_scale=1.0, noise_map=noise_map, psf=psf)


@pytest.fixture(name='mask')
def test_mask():
    return msk.Mask.circular(shape=((3, 3)), pixel_scale=0.1, radius_arcsec=0.1)


@pytest.fixture(name='fitting_image')
def test_fitting_image(image, mask):
    return fitting_data.FittingImage(image=image, mask=mask)


@pytest.fixture(name='fit_normal')
def test_fit(fitting_image):
    return fitting.AbstractImageFit(fitting_images=[fitting_image], model_images_=[np.ones(5)])


@pytest.fixture(name='hyper')
def make_hyper():
    class Hyper(object):

        def __init__(self):
            pass

    hyper = Hyper()

    hyper.hyper_model_image = np.ones((5))
    hyper.hyper_galaxy_images = [np.ones((5))]
    hyper.hyper_minimum_values = [0.2]

    hyper_galaxy = g.HyperGalaxy(contribution_factor=4.0, noise_factor=2.0, noise_power=3.0)
    hyper.hyper_galaxy = g.Galaxy(light=lp.EllipticalSersic(intensity=1.0), hyper_galaxy=hyper_galaxy)
    return hyper


@pytest.fixture(name='fitting_hyper_image')
def test_fitting_hyper_image(image, mask, hyper):
    return fitting_data.FittingHyperImage(image=image, mask=mask, hyper_model_image=hyper.hyper_model_image,
                                          hyper_galaxy_images=hyper.hyper_galaxy_images,
                                          hyper_minimum_values=hyper.hyper_minimum_values)


@pytest.fixture(name='fit_hyper')
def test_fit_hyper(fitting_hyper_image, hyper):
    return fitting.AbstractHyperImageFit(fitting_hyper_images=[fitting_hyper_image], model_images_=[np.ones(5)],
                                         hyper_galaxies=[hyper.hyper_galaxy.hyper_galaxy])


def test__model_image_is_output(fit_normal, fitting_plotter_path):
    fitting_plotters.plot_model_image(fit=fit_normal, output_path=fitting_plotter_path, output_format='png')
    assert os.path.isfile(path=fitting_plotter_path + 'fit_model_image.png')
    os.remove(path=fitting_plotter_path + 'fit_model_image.png')


def test__residuals_is_output(fit_normal, fitting_plotter_path):
    fitting_plotters.plot_residuals(fit=fit_normal, output_path=fitting_plotter_path, output_format='png')
    assert os.path.isfile(path=fitting_plotter_path + 'fit_residuals.png')
    os.remove(path=fitting_plotter_path + 'fit_residuals.png')


def test__chi_squareds_is_output(fit_normal, fitting_plotter_path):
    fitting_plotters.plot_chi_squareds(fit=fit_normal, output_path=fitting_plotter_path, output_format='png')
    assert os.path.isfile(path=fitting_plotter_path + 'fit_chi_squareds.png')
    os.remove(path=fitting_plotter_path + 'fit_chi_squareds.png')


def test__scaled_chi_squareds_is_output(fit_hyper, fitting_plotter_path):
    fitting_plotters.plot_scaled_chi_squareds(fit=fit_hyper, output_path=fitting_plotter_path,
                                              output_format='png')
    assert os.path.isfile(path=fitting_plotter_path + 'fit_scaled_chi_squareds.png')
    os.remove(path=fitting_plotter_path + 'fit_scaled_chi_squareds.png')


def test__scaled_noise_map_is_output(fit_hyper, fitting_plotter_path):
    fitting_plotters.plot_scaled_noise_map(fit=fit_hyper, output_path=fitting_plotter_path,
                                           output_format='png')
    assert os.path.isfile(path=fitting_plotter_path + 'fit_scaled_noise_map.png')
    os.remove(path=fitting_plotter_path + 'fit_scaled_noise_map.png')



from autofit import conf
from autofit.core import non_linear as nl
from autofit.core import model_mapper as mm
from autolens.data.imaging import image as im
from autolens.model.galaxy import galaxy_model as gm
from autolens.model.inversion import pixelizations as pix, regularization as reg
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline
from autolens.data.imaging.plotters import imaging_plotters
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp

import os

# Let's go back to our complex source pipeline, but this time, as you've probably guessed, fit it with an inversion.
# As we discussed in the previous tutorial, we'll begin by modeling the source with a light profile, to initialize the
# mass model, and then switch over to an inversion.

# To setup the config and output paths without docker, you need to uncomment and run the command below. If you are
# using Docker, you don't need to do anything so leave this uncommented!
path = '{}/../../'.format(os.path.dirname(os.path.realpath(__file__)))
conf.instance = conf.Config(config_path=path+'config', output_path=path+'output')

def simulate():

    from autolens.data.array import grids
    from autolens.model.galaxy import galaxy as g
    from autolens.lensing import ray_tracing

    psf = im.PSF.simulate_as_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    image_plane_grids = grids.DataGrids.grids_for_simulation(shape=(180, 180), pixel_scale=0.05, psf_shape=(11, 11))

    lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal( centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0,
                                                         einstein_radius=1.6))
    source_galaxy_0 = g.Galaxy(light=lp.EllipticalSersic(centre=(0.1, 0.1), axis_ratio=0.8, phi=90.0, intensity=0.2,
                                                         effective_radius=1.0, sersic_index=1.5))
    source_galaxy_1 = g.Galaxy(light=lp.EllipticalSersic(centre=(-0.25, 0.25), axis_ratio=0.7, phi=45.0, intensity=0.1,
                                                         effective_radius=0.2, sersic_index=3.0))
    source_galaxy_2 = g.Galaxy(light=lp.EllipticalSersic(centre=(0.45, -0.35), axis_ratio=0.6, phi=90.0, intensity=0.03,
                                                         effective_radius=0.3, sersic_index=3.5))
    source_galaxy_3 = g.Galaxy(light=lp.EllipticalSersic(centre=(-0.05, -0.0), axis_ratio=0.9, phi=140.0, intensity=0.03,
                                                         effective_radius=0.1, sersic_index=4.0))

    tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy],
                                                 source_galaxies=[source_galaxy_0, source_galaxy_1,
                                                                  source_galaxy_2, source_galaxy_3],
                                                 image_plane_grids=[image_plane_grids])

    return im.Image.simulate(array=tracer.image_plane_image_for_simulation, pixel_scale=0.05,
                                        exposure_time=300.0, psf=psf, background_sky_level=0.1, add_noise=True)

# Lets simulate the regular we'll fit, which is the same complex source as the 3_pipelines.py tutorial.
image = simulate()
imaging_plotters.plot_image_subplot(image=image)

def make_pipeline():

    pipeline_name = 'howtolens_c4_t7_6_inversion'

    # This is te same phase 1 as the complex_source pipeline, which we saw will give a good fit to the overall
    # structur of the lensed source and provide a good mass model.

    phase1 = ph.LensSourcePlanePhase(lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal)),
                                     source_galaxies=dict(source=gm.GalaxyModel(light=lp.EllipticalSersic)),
                                     optimizer_class=nl.MultiNest, phase_name=pipeline_name + '/phase_1_initialize')

    phase1.optimizer.sampling_efficiency = 0.3
    phase1.optimizer.const_efficiency_mode = True

    # Now, in phase 2, lets pass the lens mass model to phase 2, and use it to fit the source with an inversion.
    # We can customize the inversion's priors like we do our light and mass profiles.

    class InversionPhase(ph.LensSourcePlanePhase):

        def pass_priors(self, previous_results):

            self.lens_galaxies.lens = previous_results[0].variable.lens
            self.source_galaxies.source.pixelization.shape_0 = mm.UniformPrior(lower_limit=20.0, upper_limit=40.0)
            self.source_galaxies.source.pixelization.shape_1 = mm.UniformPrior(lower_limit=20.0, upper_limit=40.0)

            # We know the regularization coefficient is going to be around 1.0 from the last tutorial, but lets use a
            # large prior range just to make sure we cover solutions for any regular.
            self.source_galaxies.source.regularization.coefficients_0 = mm.UniformPrior(lower_limit=0.0,
                                                                                        upper_limit=100.0)

    phase2 = InversionPhase(lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal)),
                           source_galaxies=dict(source=gm.GalaxyModel(pixelization=pix.Rectangular,
                                                                      regularization=reg.Constant)),
                           optimizer_class=nl.MultiNest, phase_name=pipeline_name + '/phase_2_inversion')

    phase2.optimizer.sampling_efficiency = 0.3
    phase2.optimizer.const_efficiency_mode = True

    return pipeline.PipelineImaging(pipeline_name, phase1, phase2)


pipeline_inversion = make_pipeline()
pipeline_inversion.run(image=image)


import os
import shutil
import pytest
import numpy as np

from autofit import conf
from autolens.data.array import scaled_array
from autolens.data.array import mask as msk
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from autolens.model.galaxy import galaxy as g, galaxy_fitting, galaxy_data as gd
from autolens.model.galaxy.plotters import galaxy_fitting_plotters


@pytest.fixture(name='general_config')
def test_general_config():
    general_config_path = "{}/../../../test_files/configs/plotting/".format(os.path.dirname(os.path.realpath(__file__)))
    conf.instance.general = conf.NamedConfig(general_config_path + "general.ini")


@pytest.fixture(name='galaxy_fitting_plotter_path')
def test_galaxy_fitting_plotter_setup():
    galaxy_plotter_path = "{}/../../../test_files/plotting/galaxy_fitting/".format(os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(galaxy_plotter_path):
        shutil.rmtree(galaxy_plotter_path)

    os.mkdir(galaxy_plotter_path)

    return galaxy_plotter_path

@pytest.fixture(name='galaxy')
def test_galaxy():
    return g.Galaxy(light=lp.SphericalSersic(intensity=1.0), mass=mp.SphericalIsothermal(einstein_radius=1.0))

@pytest.fixture(name='array')
def test_array():
    return scaled_array.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=3.0)

@pytest.fixture(name='mask')
def test_mask():
    return msk.Mask.circular(shape=((3,3)), pixel_scale=0.1, radius_arcsec=0.1)

@pytest.fixture(name='galaxy_data_intensities')
def test_galaxy_data_intensities(array, mask):
    return gd.GalaxyDataIntensities(array=array, noise_map=np.ones((3,3)), mask=mask, sub_grid_size=2)

@pytest.fixture(name='galaxy_data_surface_density')
def test_galaxy_data_surface_density(array, mask):
    return gd.GalaxyDataSurfaceDensity(array=array, noise_map=np.ones((3,3)), mask=mask, sub_grid_size=2)

@pytest.fixture(name='galaxy_data_potential')
def test_galaxy_data_potential(array, mask):
    return gd.GalaxyDataPotential(array=array, noise_map=np.ones((3,3)), mask=mask, sub_grid_size=2)

@pytest.fixture(name='galaxy_data_deflections_y')
def test_galaxy_data_deflections_y(array, mask):
    return gd.GalaxyDataDeflectionsY(array=array, noise_map=np.ones((3,3)), mask=mask, sub_grid_size=2)

@pytest.fixture(name='galaxy_data_deflections_x')
def test_galaxy_data_deflections_x(array, mask):
    return gd.GalaxyDataDeflectionsX(array=array, noise_map=np.ones((3,3)), mask=mask, sub_grid_size=2)

@pytest.fixture(name='fit_intensities')
def test_galaxy_fitting_intensities(galaxy_data_intensities, galaxy):
    return galaxy_fitting.GalaxyFit(galaxy_datas=[galaxy_data_intensities], model_galaxy=galaxy)

@pytest.fixture(name='fit_surface_density')
def test_galaxy_fitting_surface_density(galaxy_data_surface_density, galaxy):
    return galaxy_fitting.GalaxyFit(galaxy_datas=[galaxy_data_surface_density], model_galaxy=galaxy)

@pytest.fixture(name='fit_potential')
def test_galaxy_fitting_potential(galaxy_data_potential, galaxy):
    return galaxy_fitting.GalaxyFit(galaxy_datas=[galaxy_data_potential], model_galaxy=galaxy)

@pytest.fixture(name='fit_deflections')
def test_galaxy_fitting_deflections(galaxy_data_deflections_y, galaxy_data_deflections_x, galaxy):
    return galaxy_fitting.GalaxyFitDeflections(galaxy_datas=[galaxy_data_deflections_y, galaxy_data_deflections_x],
                                               model_galaxy=galaxy)

def test__fit_sub_plot__galaxy_intensities__output_dependent_on_config(fit_intensities, general_config,
                                                                           galaxy_fitting_plotter_path):

    galaxy_fitting_plotters.plot_single_subplot(fit=fit_intensities, should_plot_mask=True,
                                                output_path=galaxy_fitting_plotter_path, output_format='png')
    assert os.path.isfile(path=galaxy_fitting_plotter_path + 'galaxy_fit.png')
    os.remove(path=galaxy_fitting_plotter_path + 'galaxy_fit.png')

def test__fit_sub_plot__galaxy_surface_density__output_dependent_on_config(fit_surface_density, general_config,
                                                                           galaxy_fitting_plotter_path):

    galaxy_fitting_plotters.plot_single_subplot(fit=fit_surface_density, should_plot_mask=True,
                                                output_path=galaxy_fitting_plotter_path, output_format='png')
    assert os.path.isfile(path=galaxy_fitting_plotter_path + 'galaxy_fit.png')
    os.remove(path=galaxy_fitting_plotter_path + 'galaxy_fit.png')
    
def test__fit_sub_plot__galaxy_potential__output_dependent_on_config(fit_potential, general_config,
                                                                           galaxy_fitting_plotter_path):

    galaxy_fitting_plotters.plot_single_subplot(fit=fit_potential, should_plot_mask=True,
                                                output_path=galaxy_fitting_plotter_path, output_format='png')
    assert os.path.isfile(path=galaxy_fitting_plotter_path + 'galaxy_fit.png')
    os.remove(path=galaxy_fitting_plotter_path + 'galaxy_fit.png')

def test__fit_sub_plot__galaxy_deflections__output_dependent_on_config(fit_deflections,
                                                                       general_config, galaxy_fitting_plotter_path):

    galaxy_fitting_plotters.plot_deflections_subplot(fit=fit_deflections,
                                                     should_plot_mask=True, output_path=galaxy_fitting_plotter_path,
                                                     output_format='png')
    assert os.path.isfile(path=galaxy_fitting_plotter_path + 'galaxy_fit.png')
    os.remove(path=galaxy_fitting_plotter_path + 'galaxy_fit.png')


import os

from autofit.core import non_linear as nl
from autofit.core import model_mapper as mm
from autolens.model.galaxy import galaxy, galaxy_model as gm
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from test.integration import tools

dirpath = os.path.dirname(os.path.realpath(__file__))
dirpath = os.path.dirname(dirpath)
output_path = '{}/../output/sensitivity'.format(dirpath)


def pipeline():
    pipeline_name = "sensitivity"
    data_name = '/sensitivity'

    tools.reset_paths(data_name, pipeline_name, output_path)

    lens_galaxy = galaxy.Galaxy(mass=mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.6),
                                subhalo=mp.SphericalIsothermal(centre=(1.0, 1.0), einstein_radius=0.0))
    source_galaxy = galaxy.Galaxy(light=lp.SphericalSersic(centre=(0.0, 0.0), intensity=1.0, effective_radius=0.5,
                                                           sersic_index=1.0))

    try:
        tools.simulate_integration_image(data_name=data_name, pixel_scale=0.05, lens_galaxies=[lens_galaxy],
                                         source_galaxies=[source_galaxy], target_signal_to_noise=30.0)
    except OSError:
        pass

    pipeline = make_pipeline(pipeline_name=pipeline_name)
    image = tools.load_image(data_name=data_name, pixel_scale=0.05)

    results = pipeline.run(image=image)
    for result in results:
        print(result)


def make_pipeline(pipeline_name):

    class SensitivePhase(ph.SensitivityPhase):

        def pass_priors(self, previous_results):

            self.lens_galaxies.lens.mass.centre_0 = 0.0
            self.lens_galaxies.lens.mass.centre_1 = 0.0
            self.lens_galaxies.lens.mass.einstein_radius = 1.6

            self.source_galaxies.source.light.centre_0 = 0.0
            self.source_galaxies.source.light.centre_1 = 0.0
            self.source_galaxies.source.light.intensity = 1.0
            self.source_galaxies.source.light.effective_radius = 0.5
            self.source_galaxies.source.light.sersic_index = 1.0

            self.sensitive_galaxies.subhalo.mass.centre_0 = mm.GaussianPrior(mean=0.0, sigma=1.0)
            self.sensitive_galaxies.subhalo.mass.centre_1 = mm.GaussianPrior(mean=0.0, sigma=1.0)
            self.sensitive_galaxies.subhalo.mass.kappa_s = 0.1
            self.sensitive_galaxies.subhalo.mass.scale_radius = 5.0

    phase1 = SensitivePhase(lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.SphericalIsothermal)),
                            source_galaxies=dict(source=gm.GalaxyModel(light=lp.SphericalSersic)),
                            sensitive_galaxies=dict(subhalo=gm.GalaxyModel(mass=mp.SphericalNFW)),
                            optimizer_class=nl.MultiNest,  phase_name="{}/phase1".format(pipeline_name))

    return pl.PipelineImaging(pipeline_name, phase1)


if __name__ == "__main__":
    pipeline()



import os
import shutil

import pytest

from autofit import conf
from autolens.data.array import grids
from autolens.model.profiles.plotters import profile_plotters
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp


@pytest.fixture(name='general_config')
def test_general_config():
    general_config_path = "{}/../../../test_files/configs/plotting/".format(os.path.dirname(os.path.realpath(__file__)))
    conf.instance.general = conf.NamedConfig(general_config_path+"general.ini")

@pytest.fixture(name='profile_plotter_path')
def test_profile_plotter_setup():
    
    profile_plotter_path = "{}/../../../test_files/plotting/profiles/".format(os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(profile_plotter_path):
        shutil.rmtree(profile_plotter_path)

    os.mkdir(profile_plotter_path)

    return profile_plotter_path

@pytest.fixture(name='light_profile')
def test_light_profile():
    return lp.EllipticalSersic(intensity=1.0)

@pytest.fixture(name='mass_profile')
def test_mass_profile():
    return mp.SphericalIsothermal(einstein_radius=1.0)

@pytest.fixture(name='grids')
def test_grids():
    return grids.DataGrids.from_shape_and_pixel_scale(shape=(100, 100), pixel_scale=0.05, sub_grid_size=2)


def test__intensities_is_output(light_profile, grids, profile_plotter_path):

    profile_plotters.plot_intensities(light_profile=light_profile, grid=grids.regular,
                                      output_path=profile_plotter_path, output_format='png')
    assert os.path.isfile(path=profile_plotter_path+'intensities.png')
    os.remove(path=profile_plotter_path+'intensities.png')

def test__surface_density_is_output(mass_profile, grids, profile_plotter_path):

    profile_plotters.plot_surface_density(mass_profile=mass_profile, grid=grids.regular,
                                          output_path=profile_plotter_path, output_format='png')
    assert os.path.isfile(path=profile_plotter_path+'surface_density.png')
    os.remove(path=profile_plotter_path+'surface_density.png')
    
def test__potential_is_output(mass_profile, grids, profile_plotter_path):

    profile_plotters.plot_potential(mass_profile=mass_profile, grid=grids.regular,
                                    output_path=profile_plotter_path, output_format='png')
    assert os.path.isfile(path=profile_plotter_path+'potential.png')
    os.remove(path=profile_plotter_path+'potential.png')
    
def test__deflections_y_is_output(mass_profile, grids, profile_plotter_path):

    profile_plotters.plot_deflections_y(mass_profile=mass_profile, grid=grids.regular,
                                        output_path=profile_plotter_path, output_format='png')
    assert os.path.isfile(path=profile_plotter_path+'deflections_y.png')
    os.remove(path=profile_plotter_path+'deflections_y.png')

def test__deflections_x_is_output(mass_profile, grids, profile_plotter_path):

    profile_plotters.plot_deflections_x(mass_profile=mass_profile, grid=grids.regular,
                                        output_path=profile_plotter_path, output_format='png')
    assert os.path.isfile(path=profile_plotter_path+'deflections_x.png')
    os.remove(path=profile_plotter_path+'deflections_x.png')


from matplotlib import pyplot as plt

from autolens import exc
from autofit import conf
from autolens.model.galaxy import galaxy_data as gd
from autolens.data.array.plotters import plotter_util, array_plotters
from autolens.data.fitting.plotters import fitting_plotters


def plot_single_subplot(fit, should_plot_mask=True, positions=None,
                        units='arcsec', kpc_per_arcsec=None, figsize=None, aspect='equal',
                        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                        titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
                        mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                        output_path=None, output_filename='galaxy_fit', output_format='show', ignore_config=True):

    plot_galaxy_fitting_as_subplot = conf.instance.general.get('output', 'plot_galaxy_fitting_as_subplot', bool)

    if not plot_galaxy_fitting_as_subplot and ignore_config is False:
        return

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=4)

    mask = fitting_plotters.get_mask(fit=fit, should_plot_mask=should_plot_mask)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    plot_galaxy_data_array(galaxy_data=fit.galaxy_datas[0], mask=mask, positions=positions, as_subplot=True,
                           units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                           cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                           linscale=linscale,
                           cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                           titlesize=titlesize, xlabelsize=xlabelsize,
                           ylabelsize=ylabelsize, xyticksize=xyticksize,
                           grid_pointsize=grid_pointsize, position_pointsize=position_pointsize,
                           mask_pointsize=mask_pointsize,
                           output_path=output_path, output_filename=output_filename,
                           output_format=output_format)

    plt.subplot(rows, columns, 2)

    fitting_plotters.plot_model_image(fit=fit, mask=mask, positions=positions, as_subplot=True,
                                      units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                      cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                      linscale=linscale,
                                      cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                      title='Model Galaxy', titlesize=titlesize, xlabelsize=xlabelsize,
                                      ylabelsize=ylabelsize, xyticksize=xyticksize,
                                      position_pointsize=position_pointsize, mask_pointsize=mask_pointsize,
                                      output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 3)

    fitting_plotters.plot_residuals(fit=fit, mask=mask, as_subplot=True,
                                    units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                    cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                    linscale=linscale,
                                    cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                    titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                                    xyticksize=xyticksize,
                                    position_pointsize=position_pointsize, mask_pointsize=mask_pointsize,
                                    output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 4)

    fitting_plotters.plot_chi_squareds(fit=fit, mask=mask, as_subplot=True,
                                       units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                       cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                       linscale=linscale,
                                       cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                                       xyticksize=xyticksize,
                                       position_pointsize=position_pointsize, mask_pointsize=mask_pointsize,
                                       output_path=output_path, output_filename='', output_format=output_format)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                               output_format=output_format)

    plt.close()


def plot_deflections_subplot(fit, should_plot_mask=True, positions=None,
                        units='arcsec', kpc_per_arcsec=None, figsize=None, aspect='equal',
                        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                        titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
                        mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                        output_path=None, output_filename='galaxy_fit', output_format='show', ignore_config=True):

    plot_galaxy_fitting_as_subplot = conf.instance.general.get('output', 'plot_galaxy_fitting_as_subplot', bool)

    if not plot_galaxy_fitting_as_subplot and ignore_config is False:
        return

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=8)

    mask = fitting_plotters.get_mask(fit=fit, should_plot_mask=should_plot_mask)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    plot_galaxy_data_array(galaxy_data=fit.galaxy_datas[0], mask=mask, positions=positions, as_subplot=True,
                           units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                           cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                           linscale=linscale,
                           cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                           titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                           grid_pointsize=grid_pointsize, position_pointsize=position_pointsize,
                           mask_pointsize=mask_pointsize,
                           output_path=output_path, output_filename=output_filename,
                           output_format=output_format)


    plt.subplot(rows, columns, 2)

    fitting_plotters.plot_model_image(fit=fit, image_index=0, mask=mask, positions=positions,
                                      as_subplot=True,
                                      units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                      cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                      linscale=linscale,
                                      cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                      title='Model Galaxy', titlesize=titlesize, xlabelsize=xlabelsize,
                                      ylabelsize=ylabelsize, xyticksize=xyticksize,
                                      position_pointsize=position_pointsize, mask_pointsize=mask_pointsize,
                                      output_path=output_path, output_filename='', output_format=output_format)
    plt.subplot(rows, columns, 3)

    fitting_plotters.plot_residuals(fit=fit, image_index=0, mask=mask, as_subplot=True,
                                    units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                    cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                    linscale=linscale,
                                    cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                    titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                                    xyticksize=xyticksize,
                                    position_pointsize=position_pointsize, mask_pointsize=mask_pointsize,
                                    output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 4)

    fitting_plotters.plot_chi_squareds(fit=fit, image_index=0, mask=mask, as_subplot=True,
                                       units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                       cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                       linscale=linscale,
                                       cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                                       xyticksize=xyticksize,
                                       position_pointsize=position_pointsize, mask_pointsize=mask_pointsize,
                                       output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 5)

    plot_galaxy_data_array(galaxy_data=fit.galaxy_datas[1], mask=mask, positions=positions, as_subplot=True,
                           units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                           cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                           linscale=linscale,
                           cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                           titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                           grid_pointsize=grid_pointsize, position_pointsize=position_pointsize,
                           mask_pointsize=mask_pointsize,
                           output_path=output_path, output_filename=output_filename,
                           output_format=output_format)


    plt.subplot(rows, columns, 6)

    fitting_plotters.plot_model_image(fit=fit, image_index=1, mask=mask, positions=positions, as_subplot=True,
                                      units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                      cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                      linscale=linscale,
                                      cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                      title='Model Galaxy', titlesize=titlesize, xlabelsize=xlabelsize,
                                      ylabelsize=ylabelsize, xyticksize=xyticksize,
                                      position_pointsize=position_pointsize, mask_pointsize=mask_pointsize,
                                      output_path=output_path, output_filename='', output_format=output_format)
    plt.subplot(rows, columns, 7)

    fitting_plotters.plot_residuals(fit=fit, image_index=1, mask=mask, as_subplot=True,
                                    units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                    cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                    linscale=linscale,
                                    cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                    titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                                    xyticksize=xyticksize,
                                    position_pointsize=position_pointsize, mask_pointsize=mask_pointsize,
                                    output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 8)

    fitting_plotters.plot_chi_squareds(fit=fit, image_index=1, mask=mask, as_subplot=True,
                                       units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                       cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                       linscale=linscale,
                                       cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                                       xyticksize=xyticksize,
                                       position_pointsize=position_pointsize, mask_pointsize=mask_pointsize,
                                       output_path=output_path, output_filename='', output_format=output_format)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)

    plt.close()

def plot_fitting_individuals(fit, units='kpc', output_path=None, output_format='show'):

    plot_galaxy_fitting_model_image = conf.instance.general.get('output', 'plot_galaxy_fitting_model_image', bool)
    plot_galaxy_fitting_residuals = conf.instance.general.get('output', 'plot_galaxy_fitting_residuals', bool)
    plot_galaxy_fitting_chi_squareds = conf.instance.general.get('output', 'plot_galaxy_fitting_chi_squareds', bool)
    
    mask = fitting_plotters.get_mask(fit=fit, should_plot_mask=True)

    kpc_per_arcsec = fit.tracer.image_plane.kpc_per_arcsec_proper

    if plot_galaxy_fitting_model_image:

        fitting_plotters.plot_model_image(fit=fit, mask=mask, units=units, kpc_per_arcsec=kpc_per_arcsec,
                                          output_path=output_path, output_format=output_format)

    if plot_galaxy_fitting_residuals:

        fitting_plotters.plot_residuals(fit=fit, mask=mask, units=units, kpc_per_arcsec=kpc_per_arcsec,
                                        output_path=output_path, output_format=output_format)

    if plot_galaxy_fitting_chi_squareds:

        fitting_plotters.plot_chi_squareds(fit=fit, mask=mask, units=units, kpc_per_arcsec=kpc_per_arcsec,
                                           output_path=output_path, output_format=output_format)

def plot_galaxy_data_array(galaxy_data, mask=None, positions=None, as_subplot=False,
                           units='arcsec', kpc_per_arcsec=None, figsize=None, aspect='equal',
                           cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                           cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                           titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
                           mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                           output_path=None, output_filename='galaxy_fit', output_format='show'):

    if isinstance(galaxy_data, gd.GalaxyDataIntensities):

        title='Galaxy Data Intensities'

    elif isinstance(galaxy_data, gd.GalaxyDataSurfaceDensity):

        title='Galaxy Data Surface Density'

    elif isinstance(galaxy_data, gd.GalaxyDataPotential):

        title='Galaxy Data Potential'

    elif isinstance(galaxy_data, gd.GalaxyDataDeflectionsY):

        title='Galaxy Data Deflections (y)'

    elif isinstance(galaxy_data, gd.GalaxyDataDeflectionsX):

        title='Galaxy Data Deflections (x)'

    else:

        raise exc.PlottingException('The galaxy datas supplied to plot_galaxy_data_array is not a supported type')

    array_plotters.plot_array(array=galaxy_data.array, mask=mask, positions=positions, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                              grid_pointsize=grid_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)


import numpy as np
import pytest

from autolens.model.inversion.util import mapper_util
from test.mock.mock_imaging import MockSubGrid, MockGridCollection

@pytest.fixture(name="three_pixels")
def make_three_pixels():
    return np.array([[0, 0], [0, 1], [1, 0]])


@pytest.fixture(name="five_pixels")
def make_five_pixels():
    return np.array([[0, 0], [0, 1], [1, 0], [1, 1], [1, 2]])


class TestMappingMatrix:

    def test__3_image_pixels__6_pixel_pixels__sub_grid_1x1(self, three_pixels):

        sub_to_pix = np.array([0, 1, 2])
        sub_to_regular = np.array([0, 1, 2])

        grids = MockGridCollection(regular=three_pixels, sub=MockSubGrid(three_pixels, sub_to_regular,
                                                                         sub_grid_size=1))

        mapping_matrix = mapper_util.mapping_matrix_from_sub_to_pix(sub_to_pix=sub_to_pix, pixels=6,
                                                                regular_pixels=grids.regular.shape[0],
                                                                sub_to_regular=grids.sub.sub_to_regular,
                                                                sub_grid_fraction=grids.sub.sub_grid_fraction)

        assert (mapping_matrix == np.array([[1, 0, 0, 0, 0, 0],  # Image pixel 0 maps to pix pixel 0.
                                            [0, 1, 0, 0, 0, 0],  # Image pixel 1 maps to pix pixel 1.
                                            [0, 0, 1, 0, 0, 0]])).all()  # Image pixel 2 maps to pix pixel 2

    def test__5_image_pixels__8_pixel_pixels__sub_grid_1x1(self, five_pixels):

        sub_to_pix = np.array([0, 1, 2, 7, 6])
        sub_to_regular = np.array([0, 1, 2, 3, 4])

        grids = MockGridCollection(regular=five_pixels, sub=MockSubGrid(five_pixels, sub_to_regular,
                                                                        sub_grid_size=1))

        mapping_matrix = mapper_util.mapping_matrix_from_sub_to_pix(sub_to_pix=sub_to_pix, pixels=8,
                                                                regular_pixels=grids.regular.shape[0],
                                                                sub_to_regular=grids.sub.sub_to_regular,
                                                                sub_grid_fraction=grids.sub.sub_grid_fraction)
        assert (mapping_matrix == np.array(
            [[1, 0, 0, 0, 0, 0, 0, 0],  # Image image_to_pixel 0 and 3 mappers to pix pixel 0.
             [0, 1, 0, 0, 0, 0, 0, 0],  # Image image_to_pixel 1 and 4 mappers to pix pixel 1.
             [0, 0, 1, 0, 0, 0, 0, 0],
             [0, 0, 0, 0, 0, 0, 0, 1],
             [0, 0, 0, 0, 0, 0, 1, 0]])).all()  # Image image_to_pixel 2 and 5 mappers to pix pixel 2

    def test__5_image_pixels__8_pixel_pixels__sub_grid_2x2__no_overlapping_pixels(self, five_pixels):

        sub_to_pix = np.array([0, 1, 2, 3, 1, 2, 3, 4, 2, 3, 4, 5, 7, 0, 1, 3, 6, 7, 4, 2])
        sub_to_regular = np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4])

        grids = MockGridCollection(regular=five_pixels, sub=MockSubGrid(five_pixels, sub_to_regular,
                                                                        sub_grid_size=2))

        mapping_matrix = mapper_util.mapping_matrix_from_sub_to_pix(sub_to_pix=sub_to_pix, pixels=8,
                                                                regular_pixels=grids.regular.shape[0],
                                                                sub_to_regular=grids.sub.sub_to_regular,
                                                                sub_grid_fraction=grids.sub.sub_grid_fraction)

        assert (mapping_matrix == np.array(
            [[0.25, 0.25, 0.25, 0.25, 0, 0, 0, 0],
             [0, 0.25, 0.25, 0.25, 0.25, 0, 0, 0],
             [0, 0, 0.25, 0.25, 0.25, 0.25, 0, 0],
             [0.25, 0.25, 0, 0.25, 0, 0, 0, 0.25],
             [0, 0, 0.25, 0, 0.25, 0, 0.25, 0.25]])).all()

    def test__5_image_pixels__8_pixel_pixels__sub_grid_2x2__include_overlapping_pixels(self, five_pixels):

        sub_to_pix = np.array([0, 0, 0, 1, 1, 1, 0, 0, 2, 3, 4, 5, 7, 0, 1, 3, 6, 7, 4, 2])
        sub_to_regular = np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4])

        grids = MockGridCollection(regular=five_pixels, sub=MockSubGrid(five_pixels, sub_to_regular,
                                                                        sub_grid_size=2))

        mapping_matrix = mapper_util.mapping_matrix_from_sub_to_pix(sub_to_pix=sub_to_pix, pixels=8,
                                                                regular_pixels=grids.regular.shape[0],
                                                                sub_to_regular=grids.sub.sub_to_regular,
                                                                sub_grid_fraction=grids.sub.sub_grid_fraction)

        assert (mapping_matrix == np.array(
            [[0.75, 0.25, 0, 0, 0, 0, 0, 0],
             [0.5, 0.5, 0, 0, 0, 0, 0, 0],
             [0, 0, 0.25, 0.25, 0.25, 0.25, 0, 0],
             [0.25, 0.25, 0, 0.25, 0, 0, 0, 0.25],
             [0, 0, 0.25, 0, 0.25, 0, 0.25, 0.25]])).all()

    def test__3_image_pixels__6_pixel_pixels__sub_grid_4x4(self, three_pixels):

        sub_to_pix = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1,
                                        2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
                                        0, 1, 2, 3, 4, 5, 0, 1, 2, 3, 4, 5, 0, 1, 2, 3])

        sub_to_regular = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                                 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2])

        grids = MockGridCollection(regular=three_pixels, sub=MockSubGrid(three_pixels, sub_to_regular,
                                                                         sub_grid_size=4))

        mapping_matrix = mapper_util.mapping_matrix_from_sub_to_pix(sub_to_pix=sub_to_pix, pixels=6,
                                                                regular_pixels=grids.regular.shape[0],
                                                                sub_to_regular=grids.sub.sub_to_regular,
                                                                sub_grid_fraction=grids.sub.sub_grid_fraction)

        assert (mapping_matrix == np.array(
            [[0.75, 0.25, 0, 0, 0, 0],
             [0, 0, 1.0, 0, 0, 0],
             [0.1875, 0.1875, 0.1875, 0.1875, 0.125, 0.125]])).all()


from autolens.data.array import grids
from autolens.data.array.plotters import grid_plotters

# In this example, we'll create a grid of Cartesian (y,x) coordinates, representing the arc second coordinate grid of
# an observed data-set (e.g. an image).

# In AutoLens, a grid is a set of two-dimensional (y,x) coordinates (in arc-seconds) that are deflected
# and traced by a strong lensing system.

# This coordinate grid is aligned with the image we analyze, thus each coordinate on the regular grid maps to the
# centre of each image-pixel. Lets make a grid using 100 x 100 pixels, with a pixel scale (arcsecond-to-pixel
# conversion factor) of 0.05", giving us a a 5" x 5" grid.
data_grids = grids.DataGrids.from_shape_and_pixel_scale(shape=(10, 100), pixel_scale=0.05,
                                                         sub_grid_size=2)

# If we print the regular grid, we see it consists of a set of arc-second coordinates
# (noting that the difference between each coordinate is the 'pixel_scales' of 0.05" defined above)
print('(y,x) regular-pixel 1:')
print(data_grids.regular[0])
print('(y,x) regular-pixel 2:')
print(data_grids.regular[1])
print('(y,x) regular-pixel 3:')
print(data_grids.regular[2])
print('(y,x) regular-pixel 101:')
print(data_grids.regular[100])
print('etc.')

# (PyAutoLens actually flattens the NumPy array to 1D to save memory, however it consists of all
# 10000 pixel coordinates on our 2D grid)
print(data_grids.regular.shape)

# We can also plot the grid, albeit it's a fairly bland uniform grid of dots.
grid_plotters.plot_grid(grid=data_grids.regular, title='Fairly Bland Uniform Grid Of Dots')

# A quick aside - we've tried our hardest to make figures like the one above appear a good size / have no overlapping
# labels. However, this depends somewhat on the display setting of your computer. If a figure doesn't look great,
# You can customize its appearance using standard Matplotlib variables. For example, we can change the grid's appearance
# above as follows:
grid_plotters.plot_grid(grid=data_grids.regular, title='Custom Grid', xlabelsize=20, ylabelsize=20, xyticksize=20,
                        pointsize=14, figsize=(8, 4))

# You can customize every plotter in AutoLens in this way!

# You might be wondering why the data_grids are using the attribute 'regular' (e.g. data_grids.regular).

# This is because the data-grids consist of multiple grids. The 'regular' grid is the grid where each (y,x) coordinate
# is located at the centre of each pixel on the grid.

# There is also a sub-grid, which splits each pixel into a (sub_grid_size x sub_grid_size) pixel. This is
# used for evaluating light and mass profiles at a higher resolution and thus more accurately.

# We specified a sub_grid_size of 2 above, therefore we expect 4 (2 x 2) times more sub-pixels than regular-pixels
# (still flattened to 1D).
print(data_grids.regular.shape)
print(data_grids.sub.shape)

# The sub-pixels coordinates can also be printed:
print('(y,x) sub-pixel 1 (of regular pixel 1):')
print(data_grids.sub[0])
print('(y,x) sub-pixel 2 (of regular pixel 1):')
print(data_grids.sub[1])
print('(y,x) sub-pixel 3 (of regular pixel 1):')
print(data_grids.sub[2])
print('(y,x) sub-pixel 4 (of regular pixel 1):')
print(data_grids.sub[3])
print('(y,x) sub-pixel 1 (of regular pixel 2):')
print(data_grids.sub[4])
print('etc.')

# Congratulations, you've completed your first PyAutoLens tutorial! Before moving on to the next one, experiment with
# PyAutoLens by doing the following:
#
# 1) Change the pixel-scale of the grid - what happens to the Cartesian coordinates?
# 2) Change the resolution of the grid - what happens Cartesian coordinates?
# 3) Change the sub-grid size - what do you note about the pixel-scale of sub pixels?


from autolens.data.imaging import image as im
from autolens.data.array import mask as ma
from autolens.lensing import lensing_fitting
from autolens.lensing import ray_tracing
from autolens.model.galaxy import galaxy as g
from autolens.lensing import lensing_image as li
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.data.imaging.plotters import imaging_plotters
from autolens.model.profiles.plotters import profile_plotters
from autolens.model.galaxy.plotters import galaxy_plotters
from autolens.lensing.plotters import plane_plotters
from autolens.lensing.plotters import ray_tracing_plotters

# In this chapter, you've learnt how create and fit_normal strong lenses with PyAutoLens. In particular, you've learnt:

# 1) PyAutoLens uses Cartesian grids of (y,x) coordinates to perform ray-tracing.
# 2) These grids are combined with light and mass profiles to compute regular, surface-densities, potentials and
#    deflection angles.
# 3) Profiles are combined to make galaxies.
# 4) Collections of galaxies (at the same redshift) form a plane.
# 5) A tracer_without_subhalo can make an regular-plane + source-plane strong lens system.
# 6) The Universe's cosmology can be input into this tracer_without_subhalo to convert units to physical values.
# 7) The tracer_without_subhalo's regular-plane regular can be used to simulate strong lens imaging observed on a real telescope.
# 8) That this regular can be fitted, so to as quantify how well a model strong lens system represents the observed regular.

# In this summary, we'll consider how flexible the tools PyAutoLens gives you are to study every aspect of a strong
# lens system. Lets get a 'fit_normal' to a strong lens, by setting up an regular, masks, tracer_without_subhalo, etc.

path = 'path/to/AutoLens/howtolens/chapter_1_introduction' # Unfortunately, in a Jupyter notebook you have to manually specify the path to PyAutoLens and this tutorial.
path = '/home/jammy/PyCharm/Projects/AutoLens/workspace/howtolens/chapter_1_introduction'
image = im.load_imaging_from_fits(image_path=path + '/data/regular.fits',
                                  noise_map_path=path+'/data/noise_map.fits',
                                  psf_path=path + '/data/psf.fits', pixel_scale=0.1)
mask = ma.Mask.circular(shape=image.shape, pixel_scale=image.pixel_scale, radius_arcsec=3.0)
lensing_image = li.LensingImage(image=image, mask=mask)
lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0))
source_galaxy = g.Galaxy(bulge=lp.EllipticalSersic(centre=(0.1, 0.1), axis_ratio=0.8, phi=45.0,
                                                  intensity=1.0, effective_radius=1.0, sersic_index=4.0),
                         disk=lp.EllipticalSersic(centre=(0.1, 0.1), axis_ratio=0.8, phi=45.0,
                                                  intensity=1.0, effective_radius=1.0, sersic_index=1.0))
tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                             image_plane_grids=[lensing_image.grids])
fit = lensing_fitting.fit_lensing_image_with_tracer(lensing_image=lensing_image, tracer=tracer)

# The fit_normal contains our tracer_without_subhalo, which contains our planes, which contain our grids and galaxies, which contain our
# profiles:
print(fit)
print()
print(fit.tracer)
print()
print(fit.tracer.image_plane)
print()
print(fit.tracer.source_plane)
print()
print(fit.tracer.image_plane.galaxies[0])
print()
print(fit.tracer.source_plane.galaxies[0])
print()
print(fit.tracer.image_plane.galaxies[0].mass)
print()
print(fit.tracer.source_plane.galaxies[0].bulge)
print()
print(fit.tracer.source_plane.galaxies[0].disk)
print()

# Using the plotters we've used throughout this chapter, we can visualize any aspect of a fit_normal we're interested
# in. For example, if we want to plot the regular of the source model_galaxy mass profile, we can do this in a variety of
# different ways
ray_tracing_plotters.plot_image_plane_image(tracer=fit.tracer)
plane_plotters.plot_image_plane_image(plane=fit.tracer.source_plane)
galaxy_plotters.plot_intensities(galaxy=fit.tracer.source_plane.galaxies[0], grid=fit.tracer.source_plane.grids[0].regular)

# Now, its not often you'd want to write all that code just to plot the regular of the source model_galaxy!

# However, as our fit_normal and ray-tracing becomes more complex, it is useful to know how to decompose their different
# attributes to extract different things about them. For example, we made our source-model_galaxy above with two light
# profiles, a 'bulge' and 'disk. We can plot the regular-plane regular of each component individually, if we know how to
# break-up the different components of the fit_normal and tracer_without_subhalo.
profile_plotters.plot_intensities(light_profile=fit.tracer.source_plane.galaxies[0].bulge,
                                  grid=fit.tracer.source_plane.grids[0].regular, title='Bulge Image-Plane Image')
profile_plotters.plot_intensities(light_profile=fit.tracer.source_plane.galaxies[0].disk,
                                  grid=fit.tracer.source_plane.grids[0].regular, title='Disk Image-Plane Image')

# The fit_normal also has the lensing regular, so we can plot the regular using the fit_normal too, if we so desire
imaging_plotters.plot_image_subplot(image=lensing_image.image)

# And, we're done, not just with the tutorial, but the chapter!

# To end, I want to quickly talk about code-design and structure. Yeah, I know, as a scientist, you don't like code
# and certaintly don't want to think about code! However, the point is, with PyAutoLens, you don't need to!

# Think about it - throughout this chapter, we never talk about anything like it was code. We didn't refer to
# 'variables', 'parameters' and 'functions' did we? Instead, we talked about 'galaxies', 'planes' and a 'tracer_without_subhalo'. These
# are the things that, as scientists, we use to visualize a strong lens system.

# Software that abstracts the underlying code in this way follows what is called an 'object-oriented design', and it is
# our hope with PyAutoLens that we've made the way you use it (that is, in coding speak, its 'interface') intuitive.

# However, if you do enjoy code, variables, functions, and parameters, you're probably ready to take a look at the
# PyAutoLens source-code. This can be found in the 'autolens' folder. At team PyAutoLens, we take a lot of pride
# in our source-code, so I can promise you its well written, well documented and thoroughly tested (check out the
# 'test' directory if you're curious how to test code well!).

# Okay, enough self-serving praise for PyAutoLens, lets wrap up the chapter. You've learn a lot in this chapter, but
# what you haven't learnt is how to 'model' a real strong gravitational lens.

# In the real world, we've no idea what the 'correct' set of light and mass profile parameters are that will give a
# good fit_normal to a lens. Lens modeling is the process of finding the lens model which provides the best-fit_normal, and that will
# be the focus of our next set of tutorials.


import os
from os import path

import numpy as np
import pytest
from autofit import conf
from autofit.core import model_mapper as mm
from autofit.core import non_linear

from autolens.data.imaging import image as img
from autolens.data.array import grids, mask as msk
from autolens.lensing import lensing_image as li
from autolens.model.galaxy import galaxy as g, galaxy_model as gm
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from autolens.pipeline import phase as ph

pytestmark = pytest.mark.filterwarnings(
    "ignore:Using a non-tuple sequence for multidimensional indexing is deprecated; use `arr[tuple(seq)]` instead of "
    "`arr[seq]`. In the future this will be interpreted as an array index, `arr[np.array(seq)]`, which will result "
    "either in an error or a different result.")

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(scope="session", autouse=True)
def do_something():
    conf.instance = conf.Config(
        '{}/../test_files/configs/phase'.format(directory))


shape = (10, 10)


class MockAnalysis(object):

    def __init__(self, number_galaxies, value):
        self.number_galaxies = number_galaxies
        self.value = value

    # noinspection PyUnusedLocal
    def galaxy_images_for_model(self, model):
        return self.number_galaxies * [np.array([self.value])]


class MockResults(object):
    def __init__(self, model_image=None, galaxy_images=(), constant=None):
        self.model_image = model_image
        self.galaxy_images = galaxy_images
        self.constant = constant or mm.ModelInstance()
        self.variable = mm.ModelMapper()


class NLO(non_linear.NonLinearOptimizer):
    def fit(self, analysis):
        class Fitness(object):
            def __init__(self, instance_from_physical_vector, constant):
                self.result = None
                self.instance_from_physical_vector = instance_from_physical_vector
                self.constant = constant

            def __call__(self, vector):
                instance = self.instance_from_physical_vector(vector)
                instance += self.constant

                likelihood = analysis.fit(instance)
                self.result = non_linear.Result(instance, likelihood)

                # Return Chi squared
                return -2 * likelihood

        fitness_function = Fitness(self.variable.instance_from_physical_vector, self.constant)
        fitness_function(self.variable.prior_count * [0.8])

        return fitness_function.result


@pytest.fixture(name="grids")
def make_grids(lensing_image):
    return grids.DataGrids.grids_from_mask_sub_grid_size_and_psf_shape(
        lensing_image.mask, 1, lensing_image.psf.shape)


@pytest.fixture(name="phase")
def make_phase():
    return ph.LensSourcePlanePhase(optimizer_class=NLO)


@pytest.fixture(name="galaxy")
def make_galaxy():
    return g.Galaxy()


@pytest.fixture(name="galaxy_model")
def make_galaxy_model():
    return gm.GalaxyModel()


@pytest.fixture(name="image")
def make_image():
    image = img.Image(np.array(np.zeros(shape)), pixel_scale=1.0, psf=img.PSF(np.ones((3, 3)), pixel_scale=1.0),
                      noise_map=img.NoiseMap(np.ones(shape), pixel_scale=1.0))
    return image


@pytest.fixture(name="lensing_image")
def make_lensing_image():
    image = img.Image(np.array(np.zeros(shape)), pixel_scale=1.0, psf=img.PSF(np.ones((3, 3)), pixel_scale=1.0),
                      noise_map=img.NoiseMap(np.ones(shape), pixel_scale=1.0))
    mask = msk.Mask.circular(shape=shape, pixel_scale=1, radius_arcsec=3.0)
    return li.LensingImage(image, mask)


@pytest.fixture(name="results")
def make_results():
    return MockResults(np.ones(shape),
                       galaxy_images=[np.ones(shape), np.ones(shape)])


@pytest.fixture(name="results_collection")
def make_results_collection(results):
    return ph.ResultsCollection([results])


class TestAutomaticPriorPassing(object):
    def test_galaxy_model_dict(self, phase, galaxy_model):
        phase.lens_galaxies = dict(galaxy_one=galaxy_model)
        assert phase.galaxy_model_tuples == [("galaxy_one", galaxy_model)]

    def test_match_galaxy_models_by_name(self, phase, galaxy_model, galaxy):
        phase.lens_galaxies = dict(galaxy_one=galaxy_model)
        instance = mm.ModelInstance()
        instance.galaxy_one = galaxy

        assert phase.match_instance_to_models(instance) == [("galaxy_one", galaxy, galaxy_model)]

    def test_phase_property_collections(self, phase):
        assert phase.phase_property_collections == [phase.lens_galaxies, phase.source_galaxies]

    # noinspection PyUnresolvedReferences
    def test_fit_priors(self, phase, galaxy_model, galaxy):
        argument_tuples = []

        new_galaxy_model = gm.GalaxyModel()

        def fitting_function(best_fit_galaxy, initial_galaxy_model):
            argument_tuples.append((best_fit_galaxy, initial_galaxy_model))
            return new_galaxy_model

        phase.lens_galaxies = dict(galaxy_one=galaxy_model)
        assert phase.lens_galaxies.galaxy_one is not None

        instance = mm.ModelInstance()
        instance.galaxy_one = galaxy

        phase.fit_priors(instance, fitting_function)

        assert phase.lens_galaxies.galaxy_one == new_galaxy_model
        assert argument_tuples == [(galaxy, galaxy_model)]

    def test_model_instance_sum_priority(self):
        instance_1 = mm.ModelInstance()
        galaxy_1 = g.Galaxy()
        instance_1.galaxy = galaxy_1

        instance_2 = mm.ModelInstance()
        galaxy_2 = g.Galaxy()
        instance_2.galaxy = galaxy_2

        assert (instance_1 + instance_2).galaxy == galaxy_2

    # noinspection PyUnresolvedReferences
    def test_fit_priors_with_results(self, phase):
        argument_tuples = []

        galaxy_model_one = gm.GalaxyModel()
        galaxy_model_two = gm.GalaxyModel()

        new_galaxy_model_one = gm.GalaxyModel()
        new_galaxy_model_two = gm.GalaxyModel()

        new_galaxy_models = {galaxy_model_one: new_galaxy_model_one, galaxy_model_two: new_galaxy_model_two}

        def fitting_function(best_fit_galaxy, initial_galaxy_model):
            argument_tuples.append((best_fit_galaxy, initial_galaxy_model))
            return new_galaxy_models[initial_galaxy_model]

        phase.lens_galaxies = dict(galaxy_one=galaxy_model_one, galaxy_two=galaxy_model_two)

        instance_one = mm.ModelInstance()
        galaxy_one = g.Galaxy()
        instance_one.galaxy_one = galaxy_one
        instance_one.galaxy_two = g.Galaxy()
        results_one = MockResults(constant=instance_one)

        instance_two = mm.ModelInstance()
        galaxy_two = g.Galaxy()
        instance_two.galaxy_two = galaxy_two
        results_two = MockResults(constant=instance_two)

        assert phase.lens_galaxies.galaxy_one == galaxy_model_one
        assert phase.lens_galaxies.galaxy_two == galaxy_model_two

        phase.fit_priors_with_results([results_one, results_two], fitting_function)

        assert phase.lens_galaxies.galaxy_one == new_galaxy_model_one
        assert phase.lens_galaxies.galaxy_two == new_galaxy_model_two
        assert argument_tuples == [(galaxy_one, galaxy_model_one),
                                   (galaxy_two, galaxy_model_two)]


def clean_images():
    try:
        os.remove('{}/source_lens_phase/source_image_0.fits'.format(directory))
        os.remove('{}/source_lens_phase/lens_image_0.fits'.format(directory))
        os.remove('{}/source_lens_phase/model_image_0.fits'.format(directory))
    except FileNotFoundError:
        pass
    conf.instance.data_path = directory


class TestPhase(object):

    def test_set_constants(self, phase, galaxy):
        phase.lens_galaxies = [galaxy]
        assert phase.optimizer.constant.lens_galaxies == [galaxy]
        assert phase.optimizer.variable.lens_galaxies == []

    def test_set_variables(self, phase, galaxy_model):
        phase.lens_galaxies = [galaxy_model]
        assert phase.optimizer.variable.lens_galaxies == [galaxy_model]
        assert phase.optimizer.constant.lens_galaxies == []

    def test_make_analysis(self, phase, image, lensing_image):
        analysis = phase.make_analysis(image=image)
        assert analysis.last_results is None
        assert analysis.lensing_image == lensing_image

    def test_fit(self, image):
        clean_images()

        phase = ph.LensSourcePlanePhase(optimizer_class=NLO,
                                        lens_galaxies=[gm.GalaxyModel(light=lp.EllipticalSersic)],
                                        source_galaxies=[gm.GalaxyModel(light=lp.EllipticalSersic)])
        result = phase.run(image=image)
        assert isinstance(result.constant.lens_galaxies[0], g.Galaxy)
        assert isinstance(result.constant.source_galaxies[0], g.Galaxy)

    def test_customize(self, results, image):
        class MyPlanePhaseAnd(ph.LensSourcePlanePhase):
            def pass_priors(self, previous_results):
                self.lens_galaxies = previous_results.last.constant.lens_galaxies
                self.source_galaxies = previous_results.last.variable.source_galaxies

        galaxy = g.Galaxy()
        galaxy_model = gm.GalaxyModel()

        setattr(results.constant, "lens_galaxies", [galaxy])
        setattr(results.variable, "source_galaxies", [galaxy_model])

        phase = MyPlanePhaseAnd(optimizer_class=NLO)
        phase.make_analysis(image=image, previous_results=ph.ResultsCollection([results]))

        assert phase.lens_galaxies == [galaxy]
        assert phase.source_galaxies == [galaxy_model]

    def test_default_mask_function(self, phase, image):
        assert len(li.LensingImage(image, phase.mask_function(image))) == 32

    def test_duplication(self):
        phase = ph.LensSourcePlanePhase(lens_galaxies=[gm.GalaxyModel()], source_galaxies=[gm.GalaxyModel()])

        ph.LensSourcePlanePhase()

        assert phase.lens_galaxies is not None
        assert phase.source_galaxies is not None

    def test_modify_image(self, image):
        class MyPhase(ph.PhaseImaging):
            def modify_image(self, im, previous_results):
                assert image.shape == im.shape
                return im

        phase = MyPhase(phase_name='phase')
        analysis = phase.make_analysis(image)
        assert analysis.lensing_image != image

    def test_tracer_for_instance(self, image):
        lens_galaxy = g.Galaxy()
        source_galaxy = g.Galaxy()
        phase = ph.LensSourcePlanePhase(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy])
        analysis = phase.make_analysis(image)
        instance = phase.constant
        tracer = analysis.tracer_for_instance(instance)

        assert tracer.image_plane.galaxies[0] == lens_galaxy
        assert tracer.source_plane.galaxies[0] == source_galaxy

    # TODO : Need to test using results

    # def test_unmasked_model_image_for_instance(self, datas_):
    #
    #     lens_galaxy = g.Galaxy(light_profile=lp.SphericalSersic(intensity=1.0))
    #     image_padded_grid = msk.PaddedRegularGrid.unmasked_grid_from_shapes_and_pixel_scale(shape=datas_.shape,
    #                                                                                         psf_shape=datas_.psf.shape,
    #                                                                                         pixel_scale=datas_.pixel_scale)
    #     image_1d = lens_galaxy.intensities_from_grid(image_padded_grid)
    #     blurred_image_1d = image_padded_grid.convolve_array_1d_with_psf(image_1d, datas_.psf)
    #     blurred_image = image_padded_grid.scaled_array_from_array_1d(blurred_image_1d)
    #
    #     phase = ph.LensPlanePhase(lens_galaxies=[lens_galaxy])
    #     analysis = phase.make_analysis(datas_)
    #     instance = phase.constant
    #     unmasked_tracer = analysis.unmasked_tracer_for_instance(instance)
    #     unmasked_model_image = analysis.unmasked_model_image_for_tracer(unmasked_tracer)
    #
    #     assert blurred_image == pytest.approx(unmasked_model_image, 1e-4)
    #
    # def test_unmasked_model_images_of_galaxies_for_instance(self, datas_):
    #
    #     g0= g.Galaxy(light_profile=lp.SphericalSersic(intensity=1.0))
    #     g1 = g.Galaxy(light_profile=lp.SphericalSersic(intensity=2.0))
    #
    #     image_padded_grid = msk.PaddedRegularGrid.unmasked_grid_from_shapes_and_pixel_scale(shape=datas_.shape,
    #                                                                                         psf_shape=datas_.psf.shape,
    #                                                                                         pixel_scale=datas_.pixel_scale)
    #
    #     g0_image_1d = g0.intensities_from_grid(image_padded_grid)
    #     g0_blurred_image_1d = image_padded_grid.convolve_array_1d_with_psf(g0_image_1d, datas_.psf)
    #     g0_blurred_image = image_padded_grid.scaled_array_from_array_1d(g0_blurred_image_1d)
    #
    #     g1_image_1d = g1.intensities_from_grid(image_padded_grid)
    #     g1_blurred_image_1d = image_padded_grid.convolve_array_1d_with_psf(g1_image_1d, datas_.psf)
    #     g1_blurred_image = image_padded_grid.scaled_array_from_array_1d(g1_blurred_image_1d)
    #
    #     phase = ph.LensPlanePhase(lens_galaxies=[g0, g1])
    #     analysis = phase.make_analysis(datas_)
    #     instance = phase.constant
    #     unmasked_tracer = analysis.unmasked_tracer_for_instance(instance)
    #     unmasked_model_images = analysis.unmasked_model_images_of_galaxies_for_tracer(unmasked_tracer)
    #
    #     assert g0_blurred_image == pytest.approx(unmasked_model_images[0], 1e-4)
    #     assert g1_blurred_image == pytest.approx(unmasked_model_images[1], 1e-4)

    def test__phase_can_receive_list_of_galaxy_models(self):
        phase = ph.LensPlanePhase(lens_galaxies=[gm.GalaxyModel(sersic=lp.EllipticalSersic,
                                                                sis=mp.SphericalIsothermal,
                                                                variable_redshift=True),
                                                 gm.GalaxyModel(sis=mp.SphericalIsothermal,
                                                                variable_redshift=True)],
                                  optimizer_class=non_linear.MultiNest)

        instance = phase.optimizer.variable.instance_from_physical_vector(
            [0.2, 0.21, 0.22, 0.23, 0.24, 0.25, 0.8, 0.1, 0.2, 0.3,
             0.4, 0.9, 0.5, 0.7, 0.8])

        assert instance.lens_galaxies[0].sersic.centre[0] == 0.2
        assert instance.lens_galaxies[0].sis.centre[0] == 0.1
        assert instance.lens_galaxies[0].sis.centre[1] == 0.2
        assert instance.lens_galaxies[0].sis.einstein_radius == 0.3
        assert instance.lens_galaxies[0].redshift == 0.4
        assert instance.lens_galaxies[1].sis.centre[0] == 0.9
        assert instance.lens_galaxies[1].sis.centre[1] == 0.5
        assert instance.lens_galaxies[1].sis.einstein_radius == 0.7
        assert instance.lens_galaxies[1].redshift == 0.8

        class LensPlanePhase2(ph.LensPlanePhase):
            # noinspection PyUnusedLocal
            def pass_models(self, previous_results):
                self.lens_galaxies[0].sis.einstein_radius = mm.Constant(10.0)

        phase = LensPlanePhase2(lens_galaxies=[gm.GalaxyModel(sersic=lp.EllipticalSersic,
                                                              sis=mp.SphericalIsothermal,
                                                              variable_redshift=True),
                                               gm.GalaxyModel(sis=mp.SphericalIsothermal,
                                                              variable_redshift=True)],
                                optimizer_class=non_linear.MultiNest)

        # noinspection PyTypeChecker
        phase.pass_models(None)

        instance = phase.optimizer.variable.instance_from_physical_vector(
            [0.01, 0.02, 0.23, 0.04, 0.05, 0.06, 0.87, 0.1, 0.2,
             0.4, 0.5, 0.5, 0.7, 0.8])
        instance += phase.optimizer.constant

        assert instance.lens_galaxies[0].sersic.centre[0] == 0.01
        assert instance.lens_galaxies[0].sis.centre[0] == 0.1
        assert instance.lens_galaxies[0].sis.centre[1] == 0.2
        assert instance.lens_galaxies[0].sis.einstein_radius == 10.0
        assert instance.lens_galaxies[0].redshift == 0.4
        assert instance.lens_galaxies[1].sis.centre[0] == 0.5
        assert instance.lens_galaxies[1].sis.centre[1] == 0.5
        assert instance.lens_galaxies[1].sis.einstein_radius == 0.7
        assert instance.lens_galaxies[1].redshift == 0.8


# class TestPixelizedPhase(object):
#     def test_constructor(self):
#         phase = ph.PixelizedSourceLensAndPhase()
#         assert isinstance(phase.source_galaxies, gm.GalaxyModel)
#         assert phase.lens_galaxies is None


# class TestAnalysis(object):
#     def test_model_image(self, results_collection, lensing_image):
#         lensing = ph.LensPlanePhase.Analysis(results_collection, lensing_image, "analysis_path")
#         assert (results_collection[0].model_image == lensing.last_results.model_image).all()


class TestResult(object):

    # def test_hyper_galaxy_and_model_images(self):
    #
    #     lensing = MockAnalysis(number_galaxies=2, value=1.0)
    #
    # result = ph.LensSourcePlanePhase.Result(constant=mm.ModelInstance(), likelihood=1,
    # variable=mm.ModelMapper(), lensing=lensing) assert (result.image_plane_source_images[0] == np.array([
    # 1.0])).all() assert (result.image_plane_source_images[1] == np.array([1.0])).all() assert (
    # result.datas_ == np.array([2.0])).all()

    def test_results(self):
        results = ph.ResultsCollection([1, 2, 3])
        assert results == [1, 2, 3]
        assert results.last == 3
        assert results.first == 1



import numpy as np
import pytest

from autolens.data.array.util import grid_util
from autolens.data.array import mask as msk, scaled_array as sca
from autolens.model.galaxy import galaxy as g, galaxy_data as gd
from autolens.model.profiles import mass_profiles as mp
from test.mock.mock_galaxy import MockGalaxy

@pytest.fixture(name='scaled_array')
def make_scaled_array():
    return sca.ScaledSquarePixelArray(array=np.ones((4, 4)), pixel_scale=3.0)

@pytest.fixture(name="mask")
def make_mask():
    return msk.Mask(np.array([[True, True, True, True],
                              [True, False, False, True],
                              [True, False, False, True],
                              [True, True, True, True]]), pixel_scale=3.0)

@pytest.fixture(name="galaxy_data")
def make_galaxy_data(scaled_array, mask):
    noise_map = sca.ScaledSquarePixelArray(array=2.0*np.ones((4,4)), pixel_scale=3.0)
    return gd.GalaxyData(array=scaled_array, noise_map=noise_map, mask=mask)


class TestGalaxyData(object):

    def test__attributes(self, scaled_array, galaxy_data):
        assert scaled_array.pixel_scale == galaxy_data.pixel_scale

    def test__scaled_array_and_mapper(self, galaxy_data):
        assert (galaxy_data == np.ones(4)).all()
        assert (galaxy_data.array == np.ones((4,4))).all()
        assert (galaxy_data.noise_map == 2.0 * np.ones((4,4))).all()
        assert (galaxy_data.mask == np.array([[True, True, True, True],
                                              [True, False, False, True],
                                              [True, False, False, True],
                                              [True, True, True, True]])).all()

    def test__grids(self, galaxy_data):

        assert galaxy_data.grids.regular.shape == (4, 2)

        assert (galaxy_data.grids.regular == np.array([[1.5, -1.5], [1.5, 1.5],
                                                       [-1.5, -1.5], [-1.5, 1.5]])).all()
        assert (galaxy_data.grids.sub == np.array([[2.0, -2.0], [2.0, -1.0], [1.0, -2.0], [1.0, -1.0],
                                                     [2.0, 1.0], [2.0, 2.0], [1.0, 1.0], [1.0, 2.0],
                                                     [-1.0, -2.0], [-1.0, -1.0], [-2.0, -2.0], [-2.0, -1.0],
                                                     [-1.0, 1.0], [-1.0, 2.0], [-2.0, 1.0], [-2.0, 2.0]])).all()

    def test__padded_grids(self, galaxy_data):

        padded_image_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(mask=np.full((4, 4), False),
                                                                                               pixel_scales=galaxy_data.array.pixel_scales)

        assert (galaxy_data.padded_grids.regular == padded_image_util).all()
        assert galaxy_data.padded_grids.regular.image_shape == (4, 4)
        assert galaxy_data.padded_grids.regular.padded_shape == (4, 4)

        padded_sub_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
            mask=np.full((4, 4), False), pixel_scales=galaxy_data.array.pixel_scales,
            sub_grid_size=galaxy_data.grids.sub.sub_grid_size)

        assert galaxy_data.padded_grids.sub == pytest.approx(padded_sub_util, 1e-4)
        assert galaxy_data.padded_grids.sub.image_shape == (4, 4)
        assert galaxy_data.padded_grids.sub.padded_shape == (4, 4)

    def test__subtract(self, galaxy_data):
        subtracted_image = galaxy_data - np.array([1, 0, 1, 0])
        assert isinstance(subtracted_image, gd.GalaxyData)
        assert subtracted_image.pixel_scale == galaxy_data.pixel_scale

        assert subtracted_image == np.array([0, 1, 0, 1])

    def test__galaxy_data_intensities(self, scaled_array, mask):

        galaxy_data = gd.GalaxyDataIntensities(array=scaled_array, noise_map=2.0*np.ones((4,4)), mask=mask,
                                               sub_grid_size=2)

        assert scaled_array.pixel_scale == galaxy_data.pixel_scale

        assert (galaxy_data[:] == np.ones(4)).all()
        assert (galaxy_data.array == np.ones((4,4))).all()
        assert (galaxy_data.noise_map_ == 2.0 * np.ones((4))).all()
        assert (galaxy_data.mask == np.array([[True, True, True, True],
                                              [True, False, False, True],
                                              [True, False, False, True],
                                              [True, True, True, True]])).all()

        galaxy = MockGalaxy(value=1, shape=4)

        intensities = galaxy_data.profile_quantity_from_galaxy_and_sub_grid(galaxy=galaxy,
                                                                            sub_grid=galaxy_data.grids.sub)

        assert (intensities == np.array([1.0, 1.0, 1.0, 1.0])).all()

        galaxy = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=1.0))

        intensities_gal = galaxy.intensities_from_grid(grid=galaxy_data.grids.sub)
        intensities_gal = galaxy_data.grids.sub.sub_data_to_regular_data(sub_array=intensities_gal)

        intensities_gd = galaxy_data.profile_quantity_from_galaxy_and_sub_grid(galaxy=galaxy,
                                                                               sub_grid=galaxy_data.grids.sub)

        assert (intensities_gal == intensities_gd).all()

    def test__galaxy_data_surface_density(self, scaled_array, mask):

        galaxy_data = gd.GalaxyDataSurfaceDensity(array=scaled_array, noise_map=2.0*np.ones((4,4)), mask=mask,
                                                  sub_grid_size=2)

        assert scaled_array.pixel_scale == galaxy_data.pixel_scale
        assert (galaxy_data == np.ones(4)).all()
        assert (galaxy_data.array == np.ones((4,4))).all()
        assert (galaxy_data.noise_map_ == 2.0 * np.ones((4))).all()
        assert (galaxy_data.mask == np.array([[True, True, True, True],
                                              [True, False, False, True],
                                              [True, False, False, True],
                                              [True, True, True, True]])).all()

        galaxy = MockGalaxy(value=1, shape=4)

        surface_density = galaxy_data.profile_quantity_from_galaxy_and_sub_grid(galaxy=galaxy,
                                                                                sub_grid=galaxy_data.grids.sub)

        assert (surface_density == np.array([1.0, 1.0, 1.0, 1.0])).all()

        surface_density_gal = galaxy.surface_density_from_grid(grid=galaxy_data.grids.sub)
        surface_density_gal = galaxy_data.grids.sub.sub_data_to_regular_data(sub_array=surface_density_gal)

        surface_density_gd = galaxy_data.profile_quantity_from_galaxy_and_sub_grid(galaxy=galaxy,
                                                                                   sub_grid=galaxy_data.grids.sub)

        assert (surface_density_gal == surface_density_gd).all()
        
    def test__galaxy_data_potential(self, scaled_array, mask):

        galaxy_data = gd.GalaxyDataPotential(array=scaled_array, noise_map=2.0*np.ones((4,4)), mask=mask,
                                             sub_grid_size=2)

        assert scaled_array.pixel_scale == galaxy_data.pixel_scale
        assert (galaxy_data == np.ones(4)).all()
        assert (galaxy_data.array == np.ones((4,4))).all()
        assert (galaxy_data.noise_map_ == 2.0 * np.ones((4))).all()
        assert (galaxy_data.mask == np.array([[True, True, True, True],
                                              [True, False, False, True],
                                              [True, False, False, True],
                                              [True, True, True, True]])).all()

        galaxy = MockGalaxy(value=1, shape=4)

        potential = galaxy_data.profile_quantity_from_galaxy_and_sub_grid(galaxy=galaxy,
                                                                                sub_grid=galaxy_data.grids.sub)

        assert (potential == np.array([1.0, 1.0, 1.0, 1.0])).all()

        galaxy = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=1.0))

        potential_gal = galaxy.potential_from_grid(grid=galaxy_data.grids.sub)
        potential_gal = galaxy_data.grids.sub.sub_data_to_regular_data(sub_array=potential_gal)

        potential_gd = galaxy_data.profile_quantity_from_galaxy_and_sub_grid(galaxy=galaxy,
                                                                                   sub_grid=galaxy_data.grids.sub)

        assert (potential_gal == potential_gd).all()
        
    def test__galaxy_data_deflections_y(self, scaled_array, mask):

        galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(einstein_radius=1.0))

        galaxy_data = gd.GalaxyDataDeflectionsY(array=scaled_array, noise_map=2.0*np.ones((4,4)), mask=mask,
                                                sub_grid_size=2)

        assert scaled_array.pixel_scale == galaxy_data.pixel_scale
        assert (galaxy_data == np.ones(4)).all()
        assert (galaxy_data.array == np.ones((4,4))).all()
        assert (galaxy_data.noise_map_ == 2.0 * np.ones((4))).all()
        assert (galaxy_data.mask == np.array([[True, True, True, True],
                                              [True, False, False, True],
                                              [True, False, False, True],
                                              [True, True, True, True]])).all()

        galaxy = MockGalaxy(value=1, shape=4)

        deflections = galaxy_data.profile_quantity_from_galaxy_and_sub_grid(galaxy=galaxy,
                                                                            sub_grid=galaxy_data.grids.sub)

        assert (deflections[:,0] == np.array([1.0, 1.0, 1.0, 1.0])).all()
        assert (deflections[:,1] == np.array([1.0, 1.0, 1.0, 1.0])).all()

        galaxy = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=1.0))

        deflections_gal = galaxy.deflections_from_grid(grid=galaxy_data.grids.sub)
        deflections_gal = np.asarray([galaxy_data.grids.sub.sub_data_to_regular_data(deflections_gal[:, 0]),
                                      galaxy_data.grids.sub.sub_data_to_regular_data(deflections_gal[:, 1])]).T

        deflections_gd = galaxy_data.profile_quantity_from_galaxy_and_sub_grid(galaxy=galaxy,
                                                                               sub_grid=galaxy_data.grids.sub)

        assert (deflections_gal[:,0] == deflections_gd[:,0]).all()
        assert (deflections_gal[:,1] == deflections_gd[:,1]).all()

    def test__galaxy_data_deflections_x(self, scaled_array, mask):

        galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(einstein_radius=1.0))

        galaxy_data = gd.GalaxyDataDeflectionsX(array=scaled_array, noise_map=2.0*np.ones((4,4)), mask=mask,
                                                sub_grid_size=2)

        assert scaled_array.pixel_scale == galaxy_data.pixel_scale
        assert (galaxy_data == np.ones(4)).all()
        assert (galaxy_data.array == np.ones((4,4))).all()
        assert (galaxy_data.noise_map_ == 2.0 * np.ones((4))).all()
        assert (galaxy_data.mask == np.array([[True, True, True, True],
                                              [True, False, False, True],
                                              [True, False, False, True],
                                              [True, True, True, True]])).all()

        galaxy = MockGalaxy(value=1, shape=4)

        deflections = galaxy_data.profile_quantity_from_galaxy_and_sub_grid(galaxy=galaxy,
                                                                            sub_grid=galaxy_data.grids.sub)

        assert (deflections[:,0] == np.array([1.0, 1.0, 1.0, 1.0])).all()
        assert (deflections[:,1] == np.array([1.0, 1.0, 1.0, 1.0])).all()

        galaxy = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=1.0))

        deflections_gal = galaxy.deflections_from_grid(grid=galaxy_data.grids.sub)
        deflections_gal = np.asarray([galaxy_data.grids.sub.sub_data_to_regular_data(deflections_gal[:, 0]),
                                      galaxy_data.grids.sub.sub_data_to_regular_data(deflections_gal[:, 1])]).T

        deflections_gd = galaxy_data.profile_quantity_from_galaxy_and_sub_grid(galaxy=galaxy,
                                                                               sub_grid=galaxy_data.grids.sub)

        assert (deflections_gal[:,0] == deflections_gd[:,0]).all()
        assert (deflections_gal[:,1] == deflections_gd[:,1]).all()


# import os
# import shutil
#
# import pytest
#
# from autofit import conf
# from autolens.imaging import scaled_array
# from autolens.imaging import regular as im
# from autolens.imaging import masks as msk
# from autolens.profiles import light_profiles as lp
# from autolens.profiles import mass_profiles as mp
# from autolens.lensing import model_galaxy as g
# from autolens.inversion import pixelizations as pix
# from autolens.inversion import regularization as reg
# from autolens.inversion import inversions as inv
# from autolens.plotting import inversion_plotters
#
# import numpy as np
#
# @pytest.fixture(name='general_config')
# def test_general_config():
#     general_config_path = "{}/../test_files/configs/plotting/".format(os.path.dirname(os.path.realpath(__file__)))
#     conf.instance.general = conf.NamedConfig(general_config_path + "general.ini")
#
#
# @pytest.fixture(name='inversion_plotter_path')
# def test_inversion_plotter_setup():
#     galaxy_plotter_path = "{}/../test_files/plotting/inversion/".format(os.path.dirname(os.path.realpath(__file__)))
#
#     if os.path.exists(galaxy_plotter_path):
#         shutil.rmtree(galaxy_plotter_path)
#
#     os.mkdir(galaxy_plotter_path)
#
#     return galaxy_plotter_path
#
# @pytest.fixture(name='regular')
# def test_image():
#
#     regular = scaled_array.ScaledSquarePixelArray(array=np.ones((3,3)), pixel_scale=1.0)
#     noise_map_ = im.NoiseMap(array=2.0*np.ones((3,3)), pixel_scale=1.0)
#     psf = im.PSF(array=3.0*np.ones((3,3)), pixel_scale=1.0)
#
#     return im.Image(array=regular, pixel_scale=1.0, noise_map_=noise_map_, psf=psf)
#
# @pytest.fixture(name='galaxy_light')
# def test_galaxy_light():
#     return g.Galaxy(light=lp.EllipticalSersic(intensity=1.0))
#
#
# @pytest.fixture(name='galaxy_mass')
# def test_galaxy_mass():
#     return g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=1.0))
#
# @pytest.fixture(name='grids')
# def test_grids():
#     return msk.ImagingGrids.from_shape_and_pixel_scale(shape=(100, 100), pixel_scale=0.05, sub_grid_size=2)
#
# @pytest.fixture(name='rectangular_pixelization')
# def test_rectangular_pixelization():
#     return pix.Rectangular(shape=(25, 25))
#
# @pytest.fixture(name='rectangular_mapper')
# def test_rectangular_mapper(rectangular_pixelization, grids):
#     return rectangular_pixelization.mapper_from_grids(grids=grids)
#
# @pytest.fixture(name='regularization')
# def test_regularization():
#     return reg.Constant(coeffs=(1.0,))
#
# @pytest.fixture(name='rectanguar_inversion')
# def test_rectangular_inversion(rectangular_mapper, regularization):
#     return inv.Inversion(im)
#
# def test__rectangular_inversion_is_output(rectangular_inversion, inversion_plotter_path):
#
#     inversion_plotters.plot_inversion(inversion=rectangular_inversion, should_plot_centres=True, should_plot_grid=True,
#                                 image_pixels=[[0, 1, 2], [3]], source_pixels=[[1, 2], [0]],
#                                 output_path=inversion_plotter_path, output_filename='rectangular_inversion',
#                                 output_format='png')
#     assert os.path.isfile(path=inversion_plotter_path + 'rectangular_inversion.png')
#     os.remove(path=inversion_plotter_path + 'rectangular_inversion.png')


import numpy as np

from autolens.model.inversion.util import regularization_util

class Regularization(object):

    def __init__(self, coefficients=(1.0,)):
        """ Abstract base class for a regularization-scheme, which is applied to a pixelization to enforce a \
        smooth-source solution and prevent over-fitting noise in the datas_. This is achieved by computing a \
        'regularization term' - which is the sum of differences in reconstructed flux between every set of neighboring \
        pixels. This regularization term is added to the solution's chi-squared as a penalty term. This effects \
        a pixelization in the following ways:

        1) The regularization matrix (see below) is added to the curvature matrix used by the reconstructor to \
        linearly invert and fit_normal the datas_ datas. Thus, it changes the pixelization in a linear manner, ensuring that \
        the minimum chi-squared solution is achieved accounting for the penalty term.

        2) The likelihood of the pixelization's fit_normal to the datas changes from L = -0.5 *(chi^2 + _noise_term) to \
        L = -0.5 (chi^2 + coefficients * regularization_term + _noise_term). The regularization \
        coefficient is a 'hyper-parameter' which determines how strongly we enforce our smoothing of the pixelization.

        The value of the coefficients(s) is set using the Bayesian framework of (Suyu 2006) and this \
        is described further in the (*inversion.Inversion* class).

        The regularization matrix, H, is calculated by defining a set of B matrices which describe how the \
        pixels neighbor one another. For howtolens, lets take a 3x3 square grid:
        ______
        |0|1|2|
        |3|4|5|
        |6|7|8|
        ^^^^^^^

        We want to regularize this grid such that each pixel is regularized with the pixel to its right and below it \
        (provided there are pixels in that direction). This means that:

        - pixel 0 is regularized with pixel 1 (to the right) and pixel 3 (below).
        - pixel 1 is regularized with pixel 2 (to the right) and pixel 4 (below),
        - Pixel 2 is only regularized with pixel 5, as there is no pixel to its right.
        - and so on.

        We make two 9 x 9 B matrices, which describe regularization in each direction (i.e. rightwards and downwards). \
        We simply put a -1 and 1 in each row of a pixel index where it has a neighbor, where the value 1 goes in the \
        column of its neighbor's index. Thus, the B matrix describing neighboring pixels to their right looks like:

        B_x = [-1,  1,  0,  0,  0,  0,  0,  0,  0] # [0->1]
              [ 0, -1,  1,  0,  0,  0,  0,  0,  0] # [1->2]
              [ 0,  0, -1,  0,  0,  0,  0,  0,  0] # [] NOTE - no pixel neighbor.
              [ 0,  0,  0, -1,  1,  0,  0,  0,  0] # [3->4]
              [ 0,  0,  0,  0, -1,  1,  0,  0,  0] # [4->5]
              [ 0,  0,  0,  0,  0, -1,  0,  0,  0] # [] NOTE - no pixel neighbor.
              [ 0,  0,  0,  0,  0,  0, -1,  1,  0] # [6->7]
              [ 0,  0,  0,  0,  0,  0,  0, -1,  1] # [7->8]
              [ 0,  0,  0,  0,  0,  0,  0,  0, -1] # [] NOTE - no pixel neighbor.

        We now make another B matrix for the regularization downwards:

        B_y = [-1,  0,  0,  1,  0,  0,  0,  0,  0] # [0->3]
              [ 0, -1,  0,  0,  1,  0,  0,  0,  0] # [1->4]
              [ 0,  0, -1,  0,  0,  1,  0,  0,  0] # [2->5]
              [ 0,  0,  0, -1,  0,  0,  1,  0,  0] # [3->6]
              [ 0,  0,  0,  0, -1,  0,  0,  1,  0] # [4->7]
              [ 0,  0,  0,  0,  0, -1,  0,  0,  1] # [5->8]
              [ 0,  0,  0,  0,  0,  0, -1,  0,  0] # [] NOTE - no pixel neighbor.
              [ 0,  0,  0,  0,  0,  0,  0, -1,  0] # [] NOTE - no pixel neighbor.
              [ 0,  0,  0,  0,  0,  0,  0,  0, -1] # [] NOTE - no pixel neighbor.

        After making the B matrices that represent our pixel neighbors, we can compute the regularization matrix, H, \
        of each direction as H = B * B.T (matrix multiplication).

        E.g.

        H_x = B_x.T, * B_x
        H_y = B_y.T * B_y
        H = H_x + H_y

        Whilst the howtolens above used a square-grid with regularization to the right and downwards, this matrix \
        formalism can be extended to describe regularization in more directions (e.g. upwards, to the left).

        It can also describe irregular grids, e.g. an irregular Voronoi grid, where a B matrix is computed for every \
        shared Voronoi vertex of each Voronoi pixel. The number of B matrices is now equal to the number of Voronoi \
        vertices in the pixel with the most Voronoi vertices. However, we describe below a scheme to compute this \
        solution more efficiently.

        ### COMBINING B MATRICES ###

        The B matrices above each had the -1's going down the diagonal. This is not necessary, and it is valid to put \
        each pixel pairing anywhere. So, if we had a 4x4 B matrix, where:

        - pixel 0 regularizes with pixel 1
        - pixel 2 regularizes with pixel 3
        - pixel 3 regularizes with pixel 0

        We can still set this up as one matrix (even though the pixel 0 comes up twice):

        B = [-1, 1,  0 , 0] # [0->1]
            [ 0, 0,  0 , 0] # We can skip rows by making them all zeros.
            [ 0, 0, -1 , 1] # [2->3]
            [ 1, 0,  0 ,-1] # [3->0] This is valid!

        So, for a Voronoi pixelzation, we don't have to make the same number of B matrices as Voronoi vertices,  \
        we can combine them into fewer B matrices as above.

        # SKIPPING THE B MATRIX CALCULATION #

        Infact, going through the rigmarole of computing and multiplying B matrices liek this is uncessary. It is \
        more computationally efficiently to directly compute H. This is possible, provided you know know all of the \
        neighboring pixel pairs (which, by definition, you need to know to set up the B matrices anyway). Thus, the \
       'regularization_matrix_from_pixel_neighbors' functions in this module directly compute H from the pixel
       neighbors.

        # POSITIVE DEFINITE MATRIX #

        The regularization matrix must be positive-definite, as the Bayesian framework of Suyu 2006 requires that we \
        use its determinant in the calculation.

        # CONSTANT REGULARIZATION #

        For the constant regularization_matrix scheme, there is only 1 regularization coefficient that is applied to \
        all neighboring pixels. This means that we when write B, we only need to regularize pixels in one direction \
        (e.g. pixel 0 regularizes pixel 1, but NOT visa versa). For howtolens:

        B = [-1, 1]  [0->1]
            [0, -1]  1 does not regularization with 0

        # WEIGHTED REGULARIZATION #

        For the weighted regularization scheme, each pixel is given an 'effective regularization weight', which is \
        applied when each set of pixel neighbors are regularized with one another. The motivation of this is that \
        different regions of a pixelization require different levels of regularization (e.g., high smoothing where the \
        no signal is present and less smoothing where it is, see (Nightingale, Dye and Massey 2018)).

        Unlike the constant regularization_matrix scheme, neighboring pixels must now be regularized with one another \
        in both directions (e.g. if pixel 0 regularizes pixel 1, pixel 1 must also regularize pixel 0). For howtolens: \

        B = [-1, 1]  [0->1]
            [-1, -1]  1 now also regularizes 0

        For a constant regularization coefficient this would NOT produce a positive-definite matrix. However, for
        the weighted scheme, it does!

        The regularize weights change the B matrix as shown below - we simply multiply each pixel's effective \
        regularization weight by each row of B it has a -1 in, so:

        regularization_weights = [1, 2, 3, 4]

        B = [-1, 1, 0 ,0] # [0->1]
            [0, -2, 2 ,0] # [1->2]
            [0, 0, -3 ,3] # [2->3]
            [4, 0, 0 ,-4] # [3->0]

        If our -1's werent down the diagonal this would look like:

        B = [4, 0, 0 ,-4] # [3->0]
            [0, -2, 2 ,0] # [1->2]
            [-1, 1, 0 ,0] # [0->1]
            [0, 0, -3 ,3] # [2->3] This is valid!

        Parameters
        -----------
        shape : (int, int)
            The dimensions of the rectangular grid of pixels (x_pixels, y_pixel)
        coefficients : (float,)
            The regularization_matrix coefficients used to smooth the pix reconstructed_image.            
            
        """
        self.coefficients = coefficients

    def regularization_matrix_from_pixel_neighbors(self, pixel_neighbors, pixel_neighbors_size):
        raise NotImplementedError("regularization_matrix_from_pixel_neighbors should be overridden")


class Constant(Regularization):

    def __init__(self, coefficients=(1.0,)):

        super(Constant, self).__init__(coefficients)

    def regularization_matrix_from_pixel_neighbors(self, pixel_neighbors, pixel_neighbors_size):
        return regularization_util.constant_regularization_matrix_from_pixel_neighbors(coefficients=self.coefficients,
               pixel_neighbors=pixel_neighbors, pixel_neighbors_size=pixel_neighbors_size)


class Weighted(Regularization):

    def __init__(self, coefficients=(1.0, 1.0), signal_scale=1.0):

        super(Weighted, self).__init__(coefficients)
        self.signal_scale = signal_scale

    def pixel_signals_from_images(self, pixels, regular_to_pix, galaxy_image):
        return regularization_util.weighted_pixel_signals_from_images(pixels=pixels, signal_scale=self.signal_scale,
                                                                      regular_to_pix=regular_to_pix,
                                                                      galaxy_image=galaxy_image)

    def regularization_weights_from_pixel_signals(self, pixel_signals):
        return regularization_util.weighted_regularization_weights_from_pixel_signals(coefficients=self.coefficients,
                                                                                      pixel_signals=pixel_signals)

    def regularization_matrix_from_pixel_neighbors(self, regularization_weights, pixel_neighbors, pixel_neighbors_size):
        return regularization_util.weighted_regularization_matrix_from_pixel_neighbors(
            regularization_weights=regularization_weights, pixel_neighbors=pixel_neighbors,
                                                 pixel_neighbors_size=pixel_neighbors_size)



import logging

import numpy as np

from autolens import exc
from autolens.data.array.util import mapping_util, array_util, mask_util
from autolens.data.array import scaled_array

logging.basicConfig()
logger = logging.getLogger(__name__)


class Mask(scaled_array.ScaledSquarePixelArray):
    """
    A masks represented by an ndarray where True is masked.
    """

    def __getitem__(self, coords):
        try:
            return super(Mask, self).__getitem__(coords)
        except IndexError:
            return True

    @classmethod
    def padded_for_shape_and_pixel_scale(cls, shape, pixel_scale):
        """
        Setup the masks such that all pixels are padded.

        Parameters
        ----------
        shape : (int, int)
            The (y,x) shape of the masks in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        """
        return cls(np.full(tuple(map(lambda d: int(d), shape)), False, dtype='bool'), pixel_scale, origin=(0.0, 0.0))

    @classmethod
    def masked_for_shape_and_pixel_scale(cls, shape, pixel_scale):
        """
        Setup the masks such that all pixels are masked.

        Parameters
        ----------
        shape : (int, int)
            The (y,x) shape of the masks in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        """
        return cls(np.full(tuple(map(lambda d: int(d), shape)), True, dtype='bool'), pixel_scale, origin=(0.0, 0.0))

    @classmethod
    def circular(cls, shape, pixel_scale, radius_arcsec, centre=(0., 0.)):
        """
        Setup the masks as a circle, using a specified arc second radius.

        Parameters
        ----------
        shape: (int, int)
            The (y,x) shape of the masks in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        radius_arcsec : float
            The radius (in arc seconds) of the circle within which pixels are not masked.
        centre: (float, float)
            The centre of the circle used to mask pixels.
        """
        mask = mask_util.mask_circular_from_shape_pixel_scale_and_radius(shape, pixel_scale, radius_arcsec,
                                                                         centre)
        return cls(mask.astype('bool'), pixel_scale, origin=(0.0, 0.0))

    @classmethod
    def circular_annular(cls, shape, pixel_scale, inner_radius_arcsec, outer_radius_arcsec, centre=(0., 0.)):
        """
        Setup the masks as an annulus, using a specified inner and outer radius in arc seconds.

        Parameters
        ----------
        shape : (int, int)
            The (y,x) shape of the masks in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        inner_radius_arcsec : float
            The radius (in arc seconds) of the inner circle outside of which pixels are not masked.
        outer_radius_arcsec : float
            The radius (in arc seconds) of the outer circle within which pixels are not masked.
        centre: (float, float)
            The origin of the masks.
        """
        mask = mask_util.mask_circular_annular_from_shape_pixel_scale_and_radii(shape, pixel_scale, inner_radius_arcsec,
                                                                                outer_radius_arcsec, centre)
        return cls(mask.astype('bool'), pixel_scale, origin=(0.0, 0.0))

    @classmethod
    def circular_anti_annular(cls, shape, pixel_scale, inner_radius_arcsec, outer_radius_arcsec, outer_radius_2_arcsec,
                              origin=(0., 0.)):
        """
        Setup the masks as an annulus, using a specified inner and outer radius in arc seconds.

        Parameters
        ----------
        shape : (int, int)
            The (y,x) shape of the masks in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        inner_radius_arcsec : float
            The radius (in arc seconds) of the inner circle inside of which pixels are not masked.
        outer_radius_arcsec : float
            The radius (in arc seconds) of the outer circle within which pixels are masked and outside of which they \
            are not.
        outer_radius_2_arcsec : float
            The radius (in arc seconds) of the second outer circle within which pixels are not masked and outside of \
            which they are.
        origin: (float, float)
            The origin of the masks.
        """
        mask = mask_util.mask_circular_anti_annular_from_shape_pixel_scale_and_radii(shape, pixel_scale, inner_radius_arcsec,
                                                                                     outer_radius_arcsec,
                                                                                     outer_radius_2_arcsec, origin)
        return cls(mask, pixel_scale)

    @classmethod
    def elliptical(cls, shape, pixel_scale, major_axis_radius_arcsec, axis_ratio, phi, centre=(0., 0.)):
        """
        Setup the masks as a ellipse, using a specified arc second major axis, axis-ratio and rotation angle phi \
        defined counter-clockwise from the positive x-axis.

        Parameters
        ----------
        shape: (int, int)
            The (y,x) shape of the masks in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        major_axis_radius_arcsec : float
            The major-axis (in arc seconds) of the ellipse within which pixels are not masked.
        axis_ratio : float
            The axis-ratio of the ellipse within which pixels are not masked.
        phi : float
            The rotation angle of the ellipse within which pixels are not masked, defined counter-clockwise from the \
             positive x-axis.
        centre: (float, float)
            The origin of the masks.
        """
        mask = mask_util.mask_elliptical_from_shape_pixel_scale_and_radius(shape, pixel_scale, major_axis_radius_arcsec,
                                                                          axis_ratio, phi, centre)
        return cls(mask.astype('bool'), pixel_scale, origin=(0.0, 0.0))

    @classmethod
    def elliptical_annular(cls, shape, pixel_scale,inner_major_axis_radius_arcsec, inner_axis_ratio, inner_phi,
                           outer_major_axis_radius_arcsec, outer_axis_ratio, outer_phi, centre=(0.0, 0.0)):
        """
        Setup the masks as a ellipse, using a specified arc second major axis, axis-ratio and rotation angle phi \
        defined counter-clockwise from the positive x-axis.

        Parameters
        ----------
        shape: (int, int)
            The (y,x) shape of the masks in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        inner_major_axis_radius_arcsec : float
            The major-axis (in arc seconds) of the inner ellipse within which pixels are masked.
        inner_axis_ratio : float
            The axis-ratio of the inner ellipse within which pixels are masked.
        inner_phi : float
            The rotation angle of the inner ellipse within which pixels are masked, defined counter-clockwise from the \
            positive x-axis.
        outer_major_axis_radius_arcsec : float
            The major-axis (in arc seconds) of the outer ellipse within which pixels are not masked.
        outer_axis_ratio : float
            The axis-ratio of the outer ellipse within which pixels are not masked.
        outer_phi : float
            The rotation angle of the outer ellipse within which pixels are not masked, defined counter-clockwise from
            the positive x-axis.
        centre: (float, float)
            The origin of the masks.
        """
        mask = mask_util.mask_elliptical_annular_from_shape_pixel_scale_and_radius(shape, pixel_scale,
                           inner_major_axis_radius_arcsec, inner_axis_ratio, inner_phi,
                           outer_major_axis_radius_arcsec, outer_axis_ratio, outer_phi, centre)
        return cls(mask.astype('bool'), pixel_scale, origin=(0.0, 0.0))

    @property
    def pixels_in_mask(self):
        return int(np.size(self) - np.sum(self))

    @property
    def masked_grid_index_to_pixel(self):
        """A 1D array of mappings between every masked pixel and its 2D pixel coordinates."""
        return mask_util.masked_grid_1d_index_to_2d_pixel_index_from_mask(self).astype('int')

    def map_2d_array_to_masked_1d_array(self, array_2d):
        """For a 2D datas-array (e.g. the datas_, noise-mappers, etc.) mappers it to a masked 1D array of values usin this masks.

        Parameters
        ----------
        array_2d : ndarray | None | float
            The datas to be mapped to a masked 1D array.
        """
        if array_2d is None or isinstance(array_2d, float):
            return array_2d
        return mapping_util.map_2d_array_to_masked_1d_array_from_array_2d_and_mask(self, array_2d)

    @array_util.Memoizer()
    def blurring_mask_for_psf_shape(self, psf_shape):
        """Compute the blurring masks, which represents all masked pixels whose light will be blurred into padded \
        pixels via PSF convolution.

        Parameters
        ----------
        psf_shape : (int, int)
           The shape of the psf which defines the blurring region (e.g. the shape of the PSF)
        """

        if psf_shape[0] % 2 == 0 or psf_shape[1] % 2 == 0:
            raise exc.MaskException("psf_size of exterior region must be odd")

        blurring_mask = mask_util.mask_blurring_from_mask_and_psf_shape(self, psf_shape)

        return Mask(blurring_mask, self.pixel_scale)

    @property
    def edge_pixels(self):
        """The indicies of the masks's edge pixels, where an edge pixel is any pixel inside the masks but on its edge \
        (next to at least one pixel with a *True* value).
        """
        return mask_util.edge_pixels_from_mask(self).astype('int')

    @property
    def border_pixels(self):
        """The indicies of the masks's edge pixels, where a border pixel is any pixel inside the masks but on an
         exterior edge (e.g. not the central pixels of an annulus mask).
        """
        return mask_util.border_pixels_from_mask(self).astype('int')


import numpy as np

from autofit.core import model_mapper
from autofit.core import non_linear
from autolens.pipeline import pipeline as pl


class MockAnalysis(object):

    def __init__(self, number_galaxies, shape, value):
        self.number_galaxies = number_galaxies
        self.shape = shape
        self.value = value

    # noinspection PyUnusedLocal
    def galaxy_images_for_model(self, model):
        return self.number_galaxies * [np.full(self.shape, self.value)]


class MockMask(object):
    pass


class DummyPhaseImaging(object):
    def __init__(self):
        self.masked_image = None
        self.positions = None
        self.previous_results = None
        self.phase_name = "dummy_phase"
        self.mask = None

    def run(self, masked_image, previous_results, mask=None):
        print(mask)
        self.masked_image = masked_image
        self.previous_results = previous_results
        self.mask = mask
        return non_linear.Result(model_mapper.ModelInstance(), 1)


class TestPassMask(object):
    def test_pass_mask(self):
        mask = MockMask()
        phase_1 = DummyPhaseImaging()
        phase_2 = DummyPhaseImaging()
        pipeline = pl.PipelineImaging("", phase_1, phase_2)
        pipeline.run(None, mask)

        assert phase_1.mask is mask
        assert phase_2.mask is mask


class TestPipelineImaging(object):
    def test_run_pipeline(self):
        phase_1 = DummyPhaseImaging()
        phase_2 = DummyPhaseImaging()
        pipeline = pl.PipelineImaging("", phase_1, phase_2)

        pipeline.run(None)

        assert len(phase_1.previous_results) == 0
        assert len(phase_2.previous_results) == 1

    def test_addition(self):
        phase_1 = DummyPhaseImaging()
        phase_2 = DummyPhaseImaging()
        phase_3 = DummyPhaseImaging()

        pipeline1 = pl.PipelineImaging("", phase_1, phase_2)
        pipeline2 = pl.PipelineImaging("", phase_3)

        assert (phase_1, phase_2, phase_3) == (pipeline1 + pipeline2).phases


class DummyPhasePositions(object):
    def __init__(self):
        self.positions = None
        self.previous_results = None
        self.pixel_scale = None
        self.phase_name = "dummy_phase"

    def run(self, positions, pixel_scale, previous_results):
        self.positions = positions
        self.pixel_scale = pixel_scale
        self.previous_results = previous_results
        return non_linear.Result(model_mapper.ModelInstance(), 1)


class TestPipelinePositions(object):
    def test_run_pipeline(self):
        phase_1 = DummyPhasePositions()
        phase_2 = DummyPhasePositions()
        pipeline = pl.PipelinePositions("", phase_1, phase_2)

        pipeline.run(None, None)

        assert len(phase_1.previous_results) == 0
        assert len(phase_2.previous_results) == 1

    def test_addition(self):
        phase_1 = DummyPhasePositions()
        phase_2 = DummyPhasePositions()
        phase_3 = DummyPhasePositions()

        pipeline1 = pl.PipelinePositions("", phase_1, phase_2)
        pipeline2 = pl.PipelinePositions("", phase_3)

        assert (phase_1, phase_2, phase_3) == (pipeline1 + pipeline2).phases



import numpy as np

from autolens.data.array import grids, mask as msk, scaled_array


class GalaxyData(scaled_array.ScaledSquarePixelArray):

    def __new__(cls, array, noise_map, mask, sub_grid_size=2):
        return np.array(mask.map_2d_array_to_masked_1d_array(array)).view(cls)

    def __init__(self, array, noise_map, mask, sub_grid_size=2):
        """ A galaxy-data is a collection of data components which are used to fit a galaxy to another galaxy. \
        This is where a component of a galaxy's light profiles (e.g. intensities) or mass profiles (e.g. surface \
        density, potential or deflection angles) are fitted to one another.

        This is primarily performed for automatic prior linking, as a means to efficiently link the priors of a galaxy \
        using one inferred parametrization of light or mass profiles to a new galaxy with a different parametrization \
        of light or mass profiles.

        This omits a number of the data components typically used when fitting an regular (e.g. the observed regular, PSF, \
        exposure time map), but still has a number of the other components (e.g. an effective noise-map, grids).

        Parameters
        ----------
        array : scaled_array.ScaledSquarePixelArray
            An array of the quantity of the galaxy that is being fitted (e.g. its intensities, surface density, etc.).
        noise_map : scaled_array.ScaledSquarePixelArray
            The noise-map used for computing the likelihood of each fit. This can be chosen arbritarily.
        mask: msk.Mask
            The 2D mask that is applied to regular data.
        sub_grid_size : int
            The size of the sub-grid used for computing the SubGrid (see imaging.mask.SubGrid).

        Attributes
        ----------
        noise_map_ : ndarray
            The masked 1D array of the noise-map
        grids : imaging.mask.DataGrids
            Grids of (y,x) Cartesian coordinates which map over the masked 1D data array's pixels (includes an \
            regular-grid, sub-grid, etc.)
        padded_grids : imaging.mask.DataGrids
            Grids of padded (y,x) Cartesian coordinates which map over the every data array's pixel in 1D and a \
            padded regioon to include edge's for accurate PSF convolution (includes an regular-grid, sub-grid, etc.)
        """

        super().__init__(array=array, pixel_scale=array.pixel_scale)

        self.array = array
        self.mask = mask
        self.noise_map = noise_map
        self.noise_map_ = mask.map_2d_array_to_masked_1d_array(array_2d=noise_map)
        self.sub_grid_size = sub_grid_size

        self.grids = grids.DataGrids.grids_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                 sub_grid_size=sub_grid_size,
                                                                                 psf_shape=(1, 1))

        self.padded_grids = grids.DataGrids.padded_grids_from_mask_sub_grid_size_and_psf_shape(
            mask=mask, sub_grid_size=sub_grid_size, psf_shape=(1, 1))

    def __array_finalize__(self, obj):
        super(GalaxyData, self).__array_finalize__(obj)
        if isinstance(obj, GalaxyData):
            self.array = obj.array
            self.mask = obj.mask
            self.noise_map = obj.noise_map
            self.noise_map_ = obj.noise_map_
            self.sub_grid_size = obj.sub_grid_size
            self.grids = obj.grids
            self.padded_grids = obj.padded_grids

    def profile_quantity_from_galaxy_and_sub_grid(self, galaxy, sub_grid):
        raise NotImplementedError()


class GalaxyDataIntensities(GalaxyData):
    """ A galaxy-data collection for fitting two galaxies based on their light-profile intensities."""

    def profile_quantity_from_galaxy_and_sub_grid(self, galaxy, sub_grid):
        """Generic function so that this galaxy-data type computes the correct profile quantity for the fitting.

        In this case, the light profile intensities are computed."""
        intensities = galaxy.intensities_from_grid(grid=sub_grid)
        return sub_grid.sub_data_to_regular_data(sub_array=intensities)


class GalaxyDataSurfaceDensity(GalaxyData):
    """ A galaxy-data collection for fitting two galaxies based on their mass-profile surface densities."""

    def profile_quantity_from_galaxy_and_sub_grid(self, galaxy, sub_grid):
        """Generic function so that this galaxy-data type computes the correct profile quantity for the fitting.

        In this case, the mass profile surface density is computed."""
        surface_density = galaxy.surface_density_from_grid(grid=sub_grid)
        return sub_grid.sub_data_to_regular_data(sub_array=surface_density)


class GalaxyDataPotential(GalaxyData):
    """ A galaxy-data collection for fitting two galaxies based on their mass-profile potentials."""

    def profile_quantity_from_galaxy_and_sub_grid(self, galaxy, sub_grid):
        """Generic function so that this galaxy-data type computes the correct profile quantity for the fitting.

        In this case, the mass profile potential is computed."""
        potential = galaxy.potential_from_grid(grid=sub_grid)
        return sub_grid.sub_data_to_regular_data(sub_array=potential)


class GalaxyDataDeflectionsY(GalaxyData):
    """ A galaxy-data collection for fitting two galaxies based on their mass-profile (y) deflection angles."""

    def profile_quantity_from_galaxy_and_sub_grid(self, galaxy, sub_grid):
        """Generic function so that this galaxy-data type computes the correct profile quantity for the fitting.

        In this case, the mass profile deflection angles are computed."""
        deflections = galaxy.deflections_from_grid(grid=sub_grid)
        return np.asarray([sub_grid.sub_data_to_regular_data(deflections[:, 0]),
                           sub_grid.sub_data_to_regular_data(deflections[:, 1])]).T


class GalaxyDataDeflectionsX(GalaxyData):
    """ A galaxy-data collection for fitting two galaxies based on their mass-profile (x) deflection angles."""

    def profile_quantity_from_galaxy_and_sub_grid(self, galaxy, sub_grid):
        """Generic function so that this galaxy-data type computes the correct profile quantity for the fitting.

        In this case, the mass profile deflection angles are computed."""
        deflections = galaxy.deflections_from_grid(grid=sub_grid)
        return np.asarray([sub_grid.sub_data_to_regular_data(deflections[:, 0]),
                           sub_grid.sub_data_to_regular_data(deflections[:, 1])]).T



from autofit import exc


class ImagingException(Exception):
    pass


class KernelException(Exception):
    pass


class MaskException(Exception):
    pass


class GalaxyException(Exception):
    pass


class RayTracingException(exc.FitException):
    pass


class PixelizationException(exc.FitException):
    pass


class InversionException(exc.FitException):
    pass


class FittingException(Exception):
    pass


class PlottingException(Exception):
    pass
